import json
import logging
import sys
import importlib.metadata
import yaml
import time
import copy
import os
import hashlib
import ipaddress

from typing import Union, Dict, Any, Tuple
from norfab.models import Result
from norfab.core.worker import NFPWorker, WorkerWatchDog, Task, Job
from norfab.core.inventory import merge_recursively
from norfab.core.exceptions import UnsupportedPluginError
from norfab.workers.nornir_worker.nornir_models import (
    GetNornirHosts,
    GetNornirHostsResponse,
)
from nornir import InitNornir
from nornir_salt.plugins.tasks import (
    netmiko_send_commands,
    scrapli_send_commands,
    napalm_send_commands,
    napalm_configure,
    netmiko_send_config,
    scrapli_send_config,
    nr_test,
    connections as nr_connections,
)
from nornir_salt.plugins.functions import (
    FFun_functions,
    FFun,
    ResultSerializer,
    HostsKeepalive,
    InventoryFun,
)
from nornir_salt.plugins.processors import (
    TestsProcessor,
    ToFileProcessor,
    DiffProcessor,
    DataProcessor,
    NorFabEventProcessor,
)
from nornir_napalm.plugins.tasks import napalm_get
from nornir_netmiko.tasks import netmiko_file_transfer
from nornir_salt.utils.pydantic_models import modelTestsProcessorSuite
from threading import Lock
from norfab.clients.shell_clients.nornir.nornir_picle_shell_cli import NorniCliInput

SERVICE = "nornir"

log = logging.getLogger(__name__)

# ----------------------------------------------------------------------
# Nornir Service watchdog class
# -----------------------------------------------------------------------


class WatchDog(WorkerWatchDog):
    """
    Class to monitor Nornir worker performance.

    Args:
        worker (Worker): The worker instance that this NornirWorker will manage.

    Attributes:
        worker (Worker): The worker instance being monitored.
        connections_idle_timeout (int): Timeout value for idle connections.
        connections_data (dict): Dictionary to store connection use timestamps.
        started_at (float): Timestamp when the watchdog was started.
        idle_connections_cleaned (int): Counter for idle connections cleaned.
        dead_connections_cleaned (int): Counter for dead connections cleaned.
        watchdog_tasks (list): List of tasks for the watchdog to run in a given order.
    """

    def __init__(self, worker):
        super().__init__(worker)
        self.worker = worker
        self.connections_idle_timeout = worker.nornir_worker_inventory.get(
            "connections_idle_timeout", None
        )
        self.connections_data = {}  # store connections use timestamps
        self.started_at = time.time()

        # stats attributes
        self.idle_connections_cleaned = 0
        self.dead_connections_cleaned = 0

        # list of tasks for watchdog to run in given order
        self.watchdog_tasks = [
            self.connections_clean,
            self.connections_keepalive,
        ]

    def stats(self) -> Dict:
        """
        Collects and returns statistics about the worker.

        Returns:
            dict: A dictionary containing the following keys:

                - runs (int): The number of runs executed by the worker.
                - timestamp (str): The current time in a human-readable format.
                - alive (int): The time in seconds since the worker started.
                - dead_connections_cleaned (int): The number of dead connections cleaned.
                - idle_connections_cleaned (int): The number of idle connections cleaned.
                - worker_ram_usage_mbyte (float): The current RAM usage of the worker in megabytes.
        """
        return {
            "runs": self.runs,
            "timestamp": time.ctime(),
            "alive": int(time.time() - self.started_at),
            "dead_connections_cleaned": self.dead_connections_cleaned,
            "idle_connections_cleaned": self.idle_connections_cleaned,
            "worker_ram_usage_mbyte": self.get_ram_usage(),
        }

    def configuration(self) -> Dict:
        """
        Returns the configuration settings for the worker.

        Returns:
            Dict: A dictionary containing the configuration settings:

                - "watchdog_interval" (int): The interval for the watchdog timer.
                - "connections_idle_timeout" (int): The timeout for idle connections.
        """
        return {
            "watchdog_interval": self.watchdog_interval,
            "connections_idle_timeout": self.connections_idle_timeout,
        }

    def connections_get(self) -> Dict:
        """
        Retrieve the current connections data.

        Returns:
            Dict: A dictionary containing the current connections data.
        """
        return {
            "connections": self.connections_data,
        }

    def connections_update(self, nr: Any, plugin: str) -> None:
        """
        Function to update connection use timestamps for each host

        Args:
            nr: Nornir object
            plugin: connection plugin name
        """
        conn_stats = {
            "last_use": None,
            "last_keepalive": None,
            "keepalive_count": 0,
        }
        for host_name in nr.inventory.hosts:
            self.connections_data.setdefault(host_name, {})
            self.connections_data[host_name].setdefault(plugin, conn_stats.copy())
            self.connections_data[host_name][plugin]["last_use"] = time.ctime()
        log.info(
            f"{self.worker.name} - updated connections use timestamps for '{plugin}'"
        )

    def connections_clean(self):
        """
        Cleans up idle connections based on the configured idle timeout.

        This method checks for connections that have been idle for longer than the
        specified `connections_idle_timeout` and disconnects them. The behavior
        varies depending on the value of `connections_idle_timeout`:

        - If `connections_idle_timeout` is None, no connections are disconnected.
        - If `connections_idle_timeout` is 0, all connections are disconnected.
        - If `connections_idle_timeout` is greater than 0, only connections that
          have been idle for longer than the specified timeout are disconnected.

        The method acquires a lock to ensure thread safety while modifying the
        connections data. It logs the disconnection actions and updates the
        `idle_connections_cleaned` counter.

        Raises:
            Exception: If an error occurs while attempting to disconnect idle connections, an error message is logged.
        """
        # dictionary keyed by plugin name and value as a list of hosts
        disconnect = {}
        if not self.worker.connections_lock.acquire(blocking=False):
            return
        try:
            # if idle timeout not set, connections don't age out
            if self.connections_idle_timeout is None:
                disconnect = {}
            # disconnect all connections for all hosts
            elif self.connections_idle_timeout == 0:
                disconnect = {"all": list(self.connections_data.keys())}
            # only disconnect aged/idle connections
            elif self.connections_idle_timeout > 0:
                for host_name, plugins in self.connections_data.items():
                    for plugin, conn_data in plugins.items():
                        last_use = time.mktime(time.strptime(conn_data["last_use"]))
                        age = time.time() - last_use
                        if age > self.connections_idle_timeout:
                            disconnect.setdefault(plugin, [])
                            disconnect[plugin].append(host_name)
            # run task to disconnect connections for aged hosts
            for plugin, hosts in disconnect.items():
                if not hosts:
                    continue
                aged_hosts = FFun(self.worker.nr, FL=hosts)
                aged_hosts.run(task=nr_connections, call="close", conn_name=plugin)
                log.debug(
                    f"{self.worker.name} watchdog, disconnected '{plugin}' "
                    f"connections for '{', '.join(hosts)}'"
                )
                self.idle_connections_cleaned += len(hosts)
                # wipe out connections data if all connection closed
                if plugin == "all":
                    self.connections_data = {}
                    break
                # remove disconnected plugin from host's connections_data
                for host in hosts:
                    self.connections_data[host].pop(plugin)
                    if not self.connections_data[host]:
                        self.connections_data.pop(host)
        except Exception as e:
            msg = f"{self.worker.name} - watchdog failed to close idle connections, error: {e}"
            log.error(msg)
        finally:
            self.worker.connections_lock.release()

    def connections_keepalive(self):
        """
        Keepalive connections and clean up dead connections if any.

        This method performs the following tasks:

        - If `connections_idle_timeout` is 0, it returns immediately without performing any actions.
        - Attempts to acquire a lock on `worker.connections_lock` to ensure thread safety.
        - Logs a debug message indicating that the keepalive process is running.
        - Uses `HostsKeepalive` to check and clean up dead connections, updating the `dead_connections_cleaned` counter.
        - Removes connections that are no longer present in the Nornir inventory.
        - Removes hosts from `connections_data` if they have no remaining connections.
        - Updates the keepalive statistics for each connection plugin, including the last keepalive time and keepalive count.
        - Logs an error message if an exception occurs during the keepalive process.
        - Releases the lock on `worker.connections_lock` in the `finally` block to ensure it is always released.

        Raises:
            Exception: If an error occurs during the keepalive process, it is logged as an error.
        """
        if self.connections_idle_timeout == 0:  # do not keepalive if idle is 0
            return
        if not self.worker.connections_lock.acquire(blocking=False):
            return
        try:
            log.debug(f"{self.worker.name} - watchdog running connections keepalive")
            stats = HostsKeepalive(self.worker.nr)
            self.dead_connections_cleaned += stats["dead_connections_cleaned"]
            # remove connections that are no longer present in Nornir inventory
            for host_name, host_connections in self.connections_data.items():
                # check if host is still in Nornir inventory
                if host_name not in self.worker.nr.inventory.hosts:
                    self.connections_data.pop(host_name, None)
                    continue
                # clean up specific connections for host
                for connection_name in list(host_connections.keys()):
                    if not self.worker.nr.inventory.hosts[host_name].connections.get(
                        connection_name
                    ):
                        self.connections_data[host_name].pop(connection_name)
            # remove host if no connections left
            for host_name in list(self.connections_data.keys()):
                if self.connections_data[host_name] == {}:
                    self.connections_data.pop(host_name)
            # update connections statistics
            for plugins in self.connections_data.values():
                for plugin in plugins.values():
                    plugin["last_keepalive"] = time.ctime()
                    plugin["keepalive_count"] += 1
        except Exception as e:
            msg = f"{self.worker.name} - watchdog HostsKeepalive check error: {e}"
            log.error(msg)
        finally:
            self.worker.connections_lock.release()


class NornirWorker(NFPWorker):
    """
    NornirWorker class for managing Nornir Service tasks.

    Args:
        inventory (str): Path to the inventory file.
        broker (str): Broker address.
        worker_name (str): Name of the worker.
        exit_event (threading.Event, optional): Event to signal worker exit. Defaults to None.
        init_done_event (threading.Event, optional): Event to signal initialization completion. Defaults to None.
        log_level (str, optional): Logging level. Defaults to None.
        log_queue (object, optional): Queue for logging. Defaults to None.

    Attributes:
        init_done_event (threading.Event): Event to signal initialization completion.
        tf_base_path (str): Base path for files folder saved using `tf` processor.
        connections_lock (threading.Lock): Lock for managing connections.
        nornir_inventory (dict): Inventory data for Nornir.
        watchdog (WatchDog): Watchdog instance for monitoring.
    """

    nr = None
    nornir_inventory = {}

    def __init__(
        self,
        inventory: str,
        broker: str,
        worker_name: str,
        exit_event=None,
        init_done_event=None,
        log_level: str = None,
        log_queue: object = None,
    ):
        super().__init__(
            inventory, broker, SERVICE, worker_name, exit_event, log_level, log_queue
        )
        self.init_done_event = init_done_event
        self.tf_base_path = os.path.join(self.base_dir, "tf")

        # misc attributes
        self.connections_lock = Lock()

        # initiate Nornir
        self.refresh_nornir(job=Job())

        # initiate watchdog
        self.watchdog = WatchDog(self)
        self.watchdog.start()

        # run startup hooks
        for f in self.inventory.hooks.get("nornir-startup", []):
            f["function"](self, *f.get("args", []), **f.get("kwargs", {}))

        if self.init_done_event is not None:
            self.init_done_event.set()

        log.info(f"{self.name} - Started")

    def worker_exit(self):
        """
        Executes all functions registered under the "nornir-exit" hook in the inventory.

        This method iterates through the list of hooks associated with the "nornir-exit"
        key in the inventory's hooks.

        For each hook, it calls the function specified in the hook, passing the current
        instance (`self`) as the first argument, followed by any additional positional
        and keyword arguments specified in the hook.
        """
        # run exit hooks
        for f in self.inventory.hooks.get("nornir-exit", []):
            f["function"](self, *f.get("args", []), **f.get("kwargs", {}))

    def init_nornir(self, inventory: dict) -> None:
        """
        Initializes the Nornir automation framework with the provided inventory.

        This method first closes any existing Nornir connections if present, optionally emitting a progress event.
        It then creates a new Nornir instance using the supplied inventory dictionary, which should contain
        configuration for logging, runner, hosts, groups, defaults, and user-defined settings.

        Args:
            inventory (dict): A dictionary containing Nornir inventory and configuration options.
        """
        # clean up existing Nornir instance
        with self.connections_lock:
            if self.nr is not None and self.nr.inventory.hosts:
                self.nr.close_connections()

            # initiate Nornir
            self.nr = InitNornir(
                logging=inventory.get("logging", {"enabled": False}),
                runner=inventory.get("runner", {}),
                inventory={
                    "plugin": "DictInventory",
                    "options": {
                        "hosts": inventory.get("hosts", {}),
                        "groups": inventory.get("groups", {}),
                        "defaults": inventory.get("defaults", {}),
                    },
                },
                user_defined=inventory.get("user_defined", {}),
            )

    def filter_hosts_and_validate(
        self, kwargs: Dict[str, Any], ret: Result
    ) -> Tuple[Any, Result]:
        """
        Helper method to filter hosts and validate results.

        Returns:
            tuple: (filtered_nornir, Result) where Result status set to
                `no_match` if no hosts matched.
        """
        self.nr.data.reset_failed_hosts()  # reset failed hosts before filtering
        filters = {k: kwargs.pop(k) for k in list(kwargs.keys()) if k in FFun_functions}
        filtered_nornir = FFun(self.nr, **filters)

        if not filtered_nornir.inventory.hosts:
            msg = (
                f"{self.name} - nothing to do, no hosts matched by filters '{filters}'"
            )
            log.debug(msg)
            ret.messages.append(msg)
            ret.status = "no_match"

        return filtered_nornir, ret

    @Task(fastapi={"methods": ["POST"]})
    def refresh_nornir(
        self,
        job: Job,
        progress: bool = False,
    ) -> Result:
        """
        Refreshes the Nornir instance by reloading the inventory from configured sources.

        This method performs the following steps:

            1. Loads the inventory configuration from the broker.
            2. If Netbox is specified in the inventory, pulls inventory data from Netbox.
            3. If Containerlab is specified in the inventory, pulls inventory data from Containerlab.
            4. Initializes the Nornir instance with the refreshed inventory.
            5. Optionally emits progress events at each stage if `progress` is True.

        Args:
            job: NorFab Job object containing relevant metadata
            progress (bool, optional): If True, emits progress events during the refresh process. Defaults to False.

        The inventory configuration is expected to be a dictionary with the following keys:

        - "logging": A dictionary specifying logging configuration (default: {"enabled": False}).
        - "runner": A dictionary specifying runner options (default: {}).
        - "hosts": A dictionary specifying host details (default: {}).
        - "groups": A dictionary specifying group details (default: {}).
        - "defaults": A dictionary specifying default values (default: {}).
        - "user_defined": A dictionary specifying user-defined options (default: {}).

        Returns:
            Result: A Result object indicating the outcome of the refresh operation.
        """
        ret = Result(task=f"{self.name}:refresh_nornir", result=True)

        # get inventory from broker
        self.nornir_worker_inventory = self.load_inventory()

        # pull Nornir inventory from Netbox
        if "netbox" in self.nornir_worker_inventory:
            self.nornir_inventory_load_netbox(job=job)
            job.event("Pulled Nornir inventory data from Netbox")

        # pull Nornir inventory from Containerlab
        if "containerlab" in self.nornir_worker_inventory:
            self.nornir_inventory_load_containerlab(
                job=job,
                **self.nornir_worker_inventory["containerlab"],
                re_init_nornir=False,
            )
            job.event("Pulled Nornir inventory data from Containerlab")

        job.event("Pulled inventories, refreshing Nornir instance")

        self.init_nornir(self.nornir_worker_inventory)

        job.event("Nornir instance refreshed")

        return ret

    @Task(fastapi={"methods": ["POST"]})
    def nornir_inventory_load_netbox(
        self,
        job: Job,
        progress: bool = False,
    ) -> Result:
        """
        Queries inventory data from Netbox Service and merges it into the Nornir inventory.

        This function checks if there is Netbox data in the inventory and retrieves
        it if available. It handles retries and timeout configurations, and ensures
        that necessary filters or devices are specified. The retrieved inventory
        data is then merged into the existing Nornir inventory.

        Args:
            job: NorFab Job object containing relevant metadata

        Logs:
            - Critical: If the inventory has no hosts, filters, or devices defined.
            - Error: If no inventory data is returned from Netbox.
            - Warning: If the Netbox instance returns no hosts data.
        """
        ret = Result(task=f"{self.name}:nornir_inventory_load_netbox", result=True)

        # form Netbox inventory load arguments
        if isinstance(self.nornir_worker_inventory.get("netbox"), dict):
            kwargs = self.nornir_worker_inventory["netbox"]
        elif self.nornir_worker_inventory.get("netbox") is True:
            kwargs = {}
        timeout = max(10, kwargs.pop("timeout", 100))

        # check if need to add devices list
        if "filters" not in kwargs and "devices" not in kwargs:
            if self.nornir_worker_inventory.get("hosts"):
                kwargs["devices"] = list(self.nornir_worker_inventory["hosts"])
            else:
                msg = f"{self.name} - inventory has no hosts, Netbox filters or devices defined"
                log.warning(msg)
                ret.result = False
                ret.messages = [msg]
                return ret

        nb_inventory_data = self.client.run_job(
            service="netbox",
            task="get_nornir_inventory",
            workers="any",
            kwargs=kwargs,
            timeout=timeout,
        )

        if nb_inventory_data is None:
            msg = f"{self.name} - Netbox get_nornir_inventory no inventory returned"
            log.error(msg)
            raise RuntimeError(msg)

        # merge Netbox inventory into Nornir inventory
        for wname, wdata in nb_inventory_data.items():
            if wdata["failed"] is False and wdata["result"].get("hosts"):
                merge_recursively(self.nornir_worker_inventory, wdata["result"])
                break
        else:
            msg = (
                f"{self.name} - Netbox worker(s) "
                f"'{', '.join(list(nb_inventory_data.keys()))}' returned no hosts data."
            )
            log.error(msg)
            job.event(msg, severity="ERROR")

        job.event("Completed processing Nornir inventory from Netbox")

        return ret

    @Task(fastapi={"methods": ["POST"]})
    def nornir_inventory_load_containerlab(
        self,
        job: Job,
        lab_name: str = None,
        groups: Union[None, list] = None,
        clab_workers: str = "all",
        use_default_credentials: bool = True,
        progress: bool = False,
        dry_run: bool = False,
        re_init_nornir: bool = True,
    ) -> Result:
        """
        Pulls the Nornir inventory from a Containerlab lab instance and merges it with the
        existing Nornir inventory.

        Args:
            job: NorFab Job object containing relevant metadata
            lab_name (str): The name of the Containerlab lab to retrieve the inventory from.
            groups (list, optional): A list of group names to include into the hosts' inventory.
            use_default_credentials (bool): Whether to use default credentials for the hosts.

        Returns:
            Result: A Result object indicating the success or failure of the operation.
                    If successful, the Nornir inventory is updated with the retrieved data.

        Notes:
            - The method retrieves inventory data from a Containerlab lab using a client job.
            - If the retrieved inventory contains host data, it is merged into the existing
              Nornir inventory using the `merge_recursively` function.
            - If no inventory or host data is returned, the method logs an error and marks
              the operation as failed.
            - After successful merging of inventory, Nornir instance is re-initialized with the
              updated inventory.
        """
        groups = groups or []
        ret = Result(
            task=f"{self.name}:nornir_inventory_load_containerlab", result=True
        )
        job.event(
            f"Pulling Containerlab '{lab_name or 'all'}' inventory from '{clab_workers}' workers"
        )

        clab_inventory_data = self.client.run_job(
            service="containerlab",
            task="get_nornir_inventory",
            workers=clab_workers,
            kwargs={
                "lab_name": lab_name,
                "groups": groups,
                "use_default_credentials": use_default_credentials,
            },
        )

        if clab_inventory_data is None:
            msg = f"{self.name} - Containerlab get_nornir_inventory no data returned"
            log.error(msg)
            raise RuntimeError(msg)

        job.event(f"Pulled Containerlab '{lab_name or 'all'}' lab inventory")

        if dry_run is True:
            ret.result = {w: r["result"] for w, r in clab_inventory_data.items()}
            return ret

        for wname, wdata in clab_inventory_data.items():
            # use inventory from first worker that returned hosts data
            if wdata["failed"] is False and wdata["result"].get("hosts"):
                merge_recursively(self.nornir_worker_inventory, wdata["result"])
                break
        else:
            msg = (
                f"{self.name} - Containerlab worker(s) '{', '.join(list(clab_inventory_data.keys()))}' "
                f"returned no hosts data for '{lab_name}' lab."
            )
            log.error(msg)
            raise RuntimeError(msg)

        job.event(
            f"Merged Containerlab '{lab_name or 'all'}' lab inventory with Nornir runtime inventory"
        )

        if re_init_nornir is True:
            self.init_nornir(self.nornir_worker_inventory)
            job.event("Nornir instance re-initialized")

        return ret

    def _add_processors(self, nr: Any, kwargs: Dict[str, Any], job: Job) -> Any:
        """
        Add various processors to the Nornir object based on the provided keyword arguments.

        Args:
            nr (Nornir): The Nornir object to which processors will be added.
            kwargs (dict): A dictionary of keyword arguments specifying which
                processors to add and their configurations.

        Keyword Args:
            tf (str, optional): Path to the file for ToFileProcessor.
            tf_skip_failed (bool, optional): Whether to skip failed tasks in ToFileProcessor.
            diff (str, optional): Configuration for DiffProcessor.
            diff_last (int, optional): Number of last diffs to keep for DiffProcessor.
            dp (list, optional): Configuration for DataProcessor.
            xml_flake (str, optional): Pattern for xml_flake function in DataProcessor.
            match (str, optional): Pattern for match function in DataProcessor.
            before (int, optional): Number of lines before match in DataProcessor.
            run_ttp (str, optional): Template for run_ttp function in DataProcessor.
            ttp_structure (str, optional): Structure for run_ttp results in DataProcessor.
            remove_tasks (bool, optional): Whether to remove tasks in DataProcessor and TestsProcessor.
            tests (list, optional): Configuration for TestsProcessor.
            subset (list, optional): Subset of tests for TestsProcessor.
            failed_only (bool, optional): Whether to include only failed tests in TestsProcessor.
            xpath (str, optional): XPath expression for DataProcessor.
            jmespath (str, optional): JMESPath expression for DataProcessor.
            iplkp (str, optional): IP lookup configuration for DataProcessor.
            ntfsm (bool, optional): Whether to use ntc-templates TextFSM parsing in DataProcessor.
            progress (bool, optional): Whether to emit progress events using NorFabEventProcessor.

        Returns:
            Nornir: The Nornir object with the added processors.
        """
        processors = []

        # extract parameters
        tf = kwargs.pop("tf", None)  # to file
        tf_skip_failed = kwargs.pop("tf_skip_failed", False)  # to file
        diff = kwargs.pop("diff", "")  # diff processor
        diff_last = kwargs.pop("diff_last", 1) if diff else None  # diff processor
        dp = kwargs.pop("dp", [])  # data processor
        xml_flake = kwargs.pop("xml_flake", "")  # data processor xml_flake function
        match = kwargs.pop("match", "")  # data processor match function
        before = kwargs.pop("before", 0)  # data processor match function
        run_ttp = kwargs.pop("run_ttp", None)  # data processor run_ttp function
        ttp_structure = kwargs.pop(
            "ttp_structure", "flat_list"
        )  # data processor run_ttp function
        remove_tasks = kwargs.pop("remove_tasks", True)  # tests and/or run_ttp
        tests = kwargs.pop("tests", None)  # tests
        subset = kwargs.pop("subset", [])  # tests
        failed_only = kwargs.pop("failed_only", False)  # tests
        groups = kwargs.pop("groups", [])  # tests
        xpath = kwargs.pop("xpath", "")  # xpath DataProcessor
        jmespath = kwargs.pop("jmespath", "")  # jmespath DataProcessor
        iplkp = kwargs.pop("iplkp", "")  # iplkp - ip lookup - DataProcessor
        ntfsm = kwargs.pop("ntfsm", False)  # ntfsm - ntc-templates TextFSM parsing
        progress = kwargs.pop(
            "progress", True
        )  # Emit progress events using NorFabEventProcessor

        # add processors if any
        if dp:
            processors.append(DataProcessor(dp))
        if iplkp:
            processors.append(
                DataProcessor(
                    [
                        {
                            "fun": "iplkp",
                            "use_dns": True if iplkp == "dns" else False,
                            "use_csv": iplkp if iplkp else False,
                        }
                    ]
                )
            )
        if xml_flake:
            processors.append(
                DataProcessor([{"fun": "xml_flake", "pattern": xml_flake}])
            )
        if xpath:
            processors.append(
                DataProcessor(
                    [{"fun": "xpath", "expr": xpath, "recover": True, "rm_ns": True}]
                )
            )
        if jmespath:
            processors.append(DataProcessor([{"fun": "jmespath", "expr": jmespath}]))
        if match:
            processors.append(
                DataProcessor([{"fun": "match", "pattern": match, "before": before}])
            )
        if run_ttp:
            processors.append(
                DataProcessor(
                    [
                        {
                            "fun": "run_ttp",
                            "template": run_ttp,
                            "res_kwargs": {"structure": ttp_structure},
                            "remove_tasks": remove_tasks,
                        }
                    ]
                )
            )
        if ntfsm:
            processors.append(DataProcessor([{"fun": "ntfsm"}]))
        if tests:
            processors.append(
                TestsProcessor(
                    tests=tests,
                    remove_tasks=remove_tasks,
                    failed_only=failed_only,
                    build_per_host_tests=True,
                    subset=subset,
                    render_tests=False,
                    groups=groups,
                )
            )
        if diff:
            processors.append(
                DiffProcessor(
                    diff=diff,
                    last=int(diff_last),
                    base_url=self.tf_base_path,
                    index=self.name,
                )
            )
        if progress:
            processors.append(NorFabEventProcessor(job=job))
        # append ToFileProcessor as the last one in the sequence
        if tf and isinstance(tf, str):
            processors.append(
                ToFileProcessor(
                    tf=tf,
                    base_url=self.tf_base_path,
                    index=self.name,
                    max_files=1000,
                    skip_failed=tf_skip_failed,
                    tf_index_lock=None,
                )
            )

        return nr.with_processors(processors)

    def load_job_data(self, job_data: Any) -> Dict:
        """
        Helper function to download job data YAML files and load it.

        Args:
            job_data (str): job data NorFab file path to download and load using YAML.

        Returns:dict
            data: The job data loaded from the YAML string.

        Raises:
            FileNotFoundError: If the job data is a URL and the file download fails.
        """
        if self.is_url(job_data):
            job_data = self.fetch_file(job_data)
            if job_data is None:
                msg = f"{self.name} - '{job_data}' job data file download failed"
                raise FileNotFoundError(msg)
            job_data = yaml.safe_load(job_data)

        return job_data

    # ----------------------------------------------------------------------
    # Nornir Service Jinja2 Filters
    # ----------------------------------------------------------------------

    def jinja2_network_hosts(self, network: str, pfxlen: bool = False) -> list:
        """
        Custom Jinja2 filter that return a list of hosts for a given IP network.

        Args:
            network (str): The network address in CIDR notation.
            pfxlen (bool, optional): If True, include the prefix length
                in the returned host addresses. Defaults to False.

        Returns:
            list: A list of host addresses as strings. If pfxlen is True,
                each address will include the prefix length.
        """
        ret = []
        ip_interface = ipaddress.ip_interface(network)
        prefixlen = ip_interface.network.prefixlen
        for ip in ip_interface.network.hosts():
            ret.append(f"{ip}/{prefixlen}" if pfxlen else str(ip))
        return ret

    def jinja2_nb_create_ip(
        self, prefix: str, device: str = None, interface: str = None, **kwargs
    ) -> str:
        """
        Jinja2 filter to get or create next available IP address from
        prefix using Netbox service.
        """
        kwargs["prefix"] = prefix
        kwargs["device"] = device
        kwargs["interface"] = interface
        reply = self.client.run_job(
            "netbox",
            "create_ip",
            kwargs=kwargs,
            workers="any",
            timeout=30,
        )
        # reply is a dict of {worker_name: results_dict}
        res = list(reply.values())[0]
        if res["failed"]:
            raise RuntimeError(res["messages"])
        return res["result"]["address"]

    def jinja2_nb_create_prefix(
        self, parent: str, description: str, prefixlen: int = 30, **kwargs
    ) -> str:
        """
        Jinja2 filter to get or create next available prefix from
        parent prefix using Netbox service.
        """
        kwargs["parent"] = parent
        kwargs["description"] = description
        kwargs["prefixlen"] = prefixlen
        reply = self.client.run_job(
            "netbox",
            "create_prefix",
            kwargs=kwargs,
            workers="any",
            timeout=30,
        )
        # reply is a dict of {worker_name: results_dict}
        res = list(reply.values())[0]
        if res["failed"]:
            raise RuntimeError(res["messages"])

        return res["result"]["prefix"]

    def jinja2_call_netbox(self, netbox_task: str) -> callable:
        """
        Returns a callable function to execute arbitrary NetBox service task.
        """

        def call_netbox(*args, **kwargs) -> dict:
            reply = self.client.run_job(
                "netbox",
                netbox_task,
                args=args,
                kwargs=kwargs,
                workers="any",
                timeout=300,
            )
            res = list(reply.values())[0]

            # check if has an error
            if res["failed"]:
                raise RuntimeError(res["messages"])

            # return result for single host only
            if len(kwargs.get("devices", [])) == 1:
                return res["result"][kwargs["devices"][0]]
            # return full results
            else:
                return res["result"]

        return call_netbox

    def add_jinja2_netbox(self) -> Dict:
        """
        Aggregates Jinja2 NetBox-related methods and functions into a dictionary
        for the ease of use within Jinja2 templates.
        """
        return {
            "get_connections": self.jinja2_call_netbox("get_connections"),
            "get_interfaces": self.jinja2_call_netbox("get_interfaces"),
            "get_circuits": self.jinja2_call_netbox("get_circuits"),
            "get_devices": self.jinja2_call_netbox("get_devices"),
            "rest": self.jinja2_call_netbox("rest"),
            "graphql": self.jinja2_call_netbox("graphql"),
            "create_ip": self.jinja2_nb_create_ip,
            "create_prefix": self.jinja2_nb_create_prefix,
        }

    def add_jinja2_filters(self) -> Dict:
        """
        Adds custom filters for use in Jinja2 templates using `|` syntaxis.

        Returns:
            dict (Dict): A dictionary where the keys are the names of the filters
                and the values are the corresponding filter functions.

                - "nb_create_ip": Method to get the next IP address.
                - "nb_create_prefix": Method to get next available prefix.
                - "network_hosts": Method to get IP network hosts.
        """
        return {
            "netbox.create_ip": self.jinja2_nb_create_ip,
            "netbox.create_prefix": self.jinja2_nb_create_prefix,
            "network_hosts": self.jinja2_network_hosts,
        }

    # ----------------------------------------------------------------------
    # Nornir Service Functions that exposed for calling
    # ----------------------------------------------------------------------

    @Task(
        fastapi={"methods": ["GET"]},
        input=GetNornirHosts,
        output=GetNornirHostsResponse,
    )
    def get_nornir_hosts(self, details: bool = False, **kwargs: dict) -> Result:
        """
        Retrieve a list of Nornir hosts managed by this worker.

        Args:
            details (bool): If True, returns detailed information about each host.
            **kwargs (dict): Hosts filters to apply when retrieving hosts.

        Returns:
            List[Dict]: A list of hosts with optional detailed information.
        """
        ret = Result(task=f"{self.name}:get_nornir_hosts", result={} if details else [])
        filtered_nornir, ret = self.filter_hosts_and_validate(kwargs, ret)
        if ret.status == "no_match":
            ret.result = None
        elif details:
            ret.result = {
                host_name: {
                    "platform": str(host.platform),
                    "hostname": str(host.hostname),
                    "port": str(host.port),
                    "groups": [str(g) for g in host.groups],
                    "username": str(host.username),
                }
                for host_name, host in filtered_nornir.inventory.hosts.items()
            }
        else:
            ret.result = list(filtered_nornir.inventory.hosts)
        return ret

    @Task(fastapi={"methods": ["GET"]})
    def get_inventory(self, **kwargs: dict) -> Result:
        """
        Retrieve running Nornir inventory for requested hosts

        Args:
            **kwargs (dict): Fx filters used to filter the inventory.

        Returns:
            Dict: A dictionary representation of the filtered inventory.
        """
        ret = Result(task=f"{self.name}:get_inventory", result={})
        filtered_nornir, ret = self.filter_hosts_and_validate(kwargs, ret)
        if ret.status != "no_match":
            ret.result = filtered_nornir.inventory.dict()
        return ret

    @Task(fastapi={"methods": ["GET"]})
    def get_version(self) -> Result:
        """
        Retrieve the versions of various libraries and system information.

        This method collects the version information of a predefined set of libraries
        and system details such as the Python version and platform. It attempts to
        import each library and fetch its version. If a library is not found, it is
        skipped.

        Returns:
            dict: a dictionary with the library names as keys and their respective
                version numbers as values. If a library is not found, its value will be
                an empty string.
        """
        libs = {
            "norfab": "",
            "scrapli": "",
            "scrapli-netconf": "",
            "scrapli-community": "",
            "paramiko": "",
            "netmiko": "",
            "napalm": "",
            "nornir": "",
            "ncclient": "",
            "nornir-netmiko": "",
            "nornir-napalm": "",
            "nornir-scrapli": "",
            "nornir-utils": "",
            "tabulate": "",
            "xmltodict": "",
            "puresnmp": "",
            "pygnmi": "",
            "pyyaml": "",
            "jmespath": "",
            "jinja2": "",
            "ttp": "",
            "nornir-salt": "",
            "lxml": "",
            "ttp-templates": "",
            "ntc-templates": "",
            "cerberus": "",
            "pydantic": "",
            "requests": "",
            "textfsm": "",
            "N2G": "",
            "dnspython": "",
            "pythonping": "",
            "python": sys.version.split(" ")[0],
            "platform": sys.platform,
        }
        # get version of packages installed
        for pkg in libs.keys():
            try:
                libs[pkg] = importlib.metadata.version(pkg)
            except importlib.metadata.PackageNotFoundError:
                pass

        return Result(result=libs)

    @Task(fastapi={"methods": ["GET"]})
    def get_watchdog_stats(self) -> Result:
        """
        Retrieve the statistics from the watchdog.

        Returns:
            Result: An object containing the statistics from the watchdog.
        """
        return Result(result=self.watchdog.stats())

    @Task(fastapi={"methods": ["GET"]})
    def get_watchdog_configuration(self) -> Result:
        """
        Retrieves the current configuration of the watchdog.

        Returns:
            Result: An object containing the watchdog configuration.
        """
        return Result(result=self.watchdog.configuration())

    @Task(fastapi={"methods": ["GET"]})
    def get_watchdog_connections(self) -> Result:
        """
        Retrieve the list of connections currently managed by watchdog.

        Returns:
            Result: An instance of the Result class containing the current
                watchdog connections.
        """
        return Result(result=self.watchdog.connections_get())

    @Task(fastapi={"methods": ["POST"]})
    def task(self, job: Job, plugin: str, **kwargs: Any) -> Result:
        """
        Execute a Nornir task plugin.

        This method dynamically imports and executes a specified Nornir task plugin,
        using the provided arguments and keyword arguments. The `plugin` attribute
        can refer to a file to fetch from a file service, which must contain a function
        named `task` that accepts a Nornir task object as the first positional argument.

        Example of a custom task function file:

        ```python
        # define connection name for RetryRunner to properly detect it
        CONNECTION_NAME = "netmiko"

        # create task function
        def task(nornir_task_object, **kwargs):
            pass
        ```

        Note:
            The `CONNECTION_NAME` must be defined within the custom task function file if
            RetryRunner is in use. Otherwise, the connection retry logic is skipped, and
            connections to all hosts are initiated simultaneously up to the number of `num_workers`.

        Args:
            job: NorFab Job object containing relevant metadata
            plugin (str): The path to the plugin function to import, or a NorFab
                URL to download a custom task or template URL that resolves to a file.
            **kwargs (Any): Additional arguments to pass to the specified task plugin.

        Notes:
            - `add_details` (bool): If True, adds task execution details to the results.
            - `to_dict` (bool): If True, returns results as a dictionary. Defaults to True.
            - Host filters: keys matching `FFun_functions` are treated as Nornir host filters.

        Returns:
            Result: An instance of the Result class containing the task execution results.

        Raises:
            FileNotFoundError: If the specified plugin file cannot be downloaded.
        """
        # extract attributes
        add_details = kwargs.pop("add_details", False)  # ResultSerializer
        to_dict = kwargs.pop("to_dict", True)  # ResultSerializer
        ret = Result(task=f"{self.name}:task", result={} if to_dict else [])

        filtered_nornir, no_match_result = self.filter_hosts_and_validate(kwargs, ret)
        if ret.status == "no_match":
            return ret

        # download task from broker and load it
        if plugin.startswith("nf://"):
            function_text = self.fetch_file(plugin)
            if function_text is None:
                raise FileNotFoundError(
                    f"{self.name} - '{plugin}' task plugin download failed"
                )

            # load task function running exec
            globals_dict = {}
            exec(function_text, globals_dict, globals_dict)
            task_function = globals_dict["task"]
        # import task function
        elif "." in plugin:
            # below same as "from nornir.plugins.tasks import task_fun as task_function"
            module_name, func_name = plugin.rsplit(".", 1)
            module = __import__(module_name, fromlist=[func_name])  # nosec
            task_function = getattr(module, func_name)
        else:
            raise RuntimeError(
                f"{self.name} - '{plugin}' task should either be a path "
                f"to a file or a module import string"
            )

        nr = self._add_processors(filtered_nornir, kwargs, job)  # add processors

        # run task
        log.debug(f"{self.name} - running Nornir task '{plugin}', kwargs '{kwargs}'")
        with self.connections_lock:
            result = nr.run(task=task_function, **kwargs)

        ret.failed = result.failed  # failed is true if any of the hosts failed
        ret.result = ResultSerializer(result, to_dict=to_dict, add_details=add_details)

        self.watchdog.connections_clean()

        return ret

    @Task(
        fastapi={"methods": ["POST"]},
        input=NorniCliInput,
    )
    def cli(
        self,
        job: Job,
        commands: Union[str, list] = None,
        plugin: str = "netmiko",
        dry_run: bool = False,
        run_ttp: str = None,
        job_data: Any = None,
        to_dict: bool = True,
        add_details: bool = False,
        **kwargs: Any,
    ) -> Result:
        """
        Task to collect/retrieve show commands output from network devices using
        Command Line Interface (CLI).

        Must either provide list of commands to run or TTP template to run.

        Args:
            job: NorFab Job object containing relevant metadata
            commands (list, optional): List of commands to send to devices or URL to a file or template
                URL that resolves to a file.
            plugin (str, optional): Plugin name to use. Valid options are
                ``netmiko``, ``scrapli``, ``napalm``.
            dry_run (bool, optional): If True, do not send commands to devices,
                just return them.
            run_ttp (str, optional): TTP Template to run.
            job_data (str, optional): URL to YAML file with data or dictionary/list
                of data to pass on to Jinja2 rendering context.
            to_dict (bool, optional): If True, returns results as a dictionary.
            add_details (bool, optional): If True, adds task execution details
                to the results.
            **kwargs: Additional arguments to pass to the specified task plugin.

        Returns:
            dict: A dictionary with the results of the CLI task.

        Raises:
            UnsupportedPluginError: If the specified plugin is not supported.
            FileNotFoundError: If the specified TTP template or job data file
                cannot be downloaded.
        """
        job_data = job_data or {}
        timeout = job.timeout * 0.9
        ret = Result(task=f"{self.name}:cli", result={} if to_dict else [])

        filtered_nornir, no_match_result = self.filter_hosts_and_validate(kwargs, ret)
        if ret.status == "no_match":
            return ret

        # decide on what send commands task plugin to use
        if plugin == "netmiko":
            task_plugin = netmiko_send_commands
            if kwargs.get("use_ps"):
                kwargs.setdefault("timeout", timeout)
            else:
                kwargs.setdefault("read_timeout", timeout)
        elif plugin == "scrapli":
            task_plugin = scrapli_send_commands
            kwargs.setdefault("timeout_ops", timeout)
        elif plugin == "napalm":
            task_plugin = napalm_send_commands
        else:
            raise UnsupportedPluginError(f"Plugin '{plugin}' not supported")

        # download TTP template
        if self.is_url(run_ttp):
            downloaded = self.fetch_file(run_ttp)
            kwargs["run_ttp"] = downloaded
            if downloaded is None:
                msg = f"{self.name} - TTP template download failed '{run_ttp}'"
                raise FileNotFoundError(msg)
        # use TTP template as is - inline template or ttp://xyz path
        elif run_ttp:
            kwargs["run_ttp"] = run_ttp

        # download job data
        job_data = self.load_job_data(job_data)

        nr = self._add_processors(filtered_nornir, kwargs, job)  # add processors

        # render commands using Jinja2 on a per-host basis
        if commands:
            commands = commands if isinstance(commands, list) else [commands]
            for host in nr.inventory.hosts.values():
                rendered = self.jinja2_render_templates(
                    templates=commands,
                    context={
                        "host": host,
                        "norfab": self.client,
                        "job_data": job_data,
                        "netbox": self.add_jinja2_netbox(),
                    },
                    filters=self.add_jinja2_filters(),
                )
                host.data["__task__"] = {"commands": rendered}

        # run task
        log.debug(
            f"{self.name} - running cli commands '{commands}', kwargs '{kwargs}', is cli dry run - '{dry_run}'"
        )
        if dry_run is True:
            result = nr.run(
                task=nr_test, use_task_data="commands", name="dry_run", **kwargs
            )
        else:
            with self.connections_lock:
                result = nr.run(task=task_plugin, **kwargs)

        ret.failed = result.failed  # failed is true if any of the hosts failed
        ret.result = ResultSerializer(result, to_dict=to_dict, add_details=add_details)

        # remove __task__ data
        for host_name, host_object in nr.inventory.hosts.items():
            _ = host_object.data.pop("__task__", None)

        self.watchdog.connections_update(nr, plugin)
        self.watchdog.connections_clean()

        return ret

    @Task(
        fastapi={"methods": ["POST"]},
    )
    def cfg(
        self,
        job: Job,
        config: Union[str, list],
        plugin: str = "netmiko",
        dry_run: bool = False,
        to_dict: bool = True,
        add_details: bool = False,
        job_data: Any = None,
        **kwargs: Any,
    ) -> Result:
        """
        Task to send configuration commands to devices using Command Line Interface (CLI).

        Args:
            job: NorFab Job object containing relevant metadata
            config (list): List of commands to send to devices or URL to a file or template
                URL that resolves to a file.
            plugin (str, optional): Plugin name to use. Valid options are:

                - netmiko - use Netmiko to configure devices
                - scrapli - use Scrapli to configure devices
                - napalm - use NAPALM to configure devices

            dry_run (bool, optional): If True, will not send commands to devices but just return them.
            to_dict (bool, optional): If True, returns results as a dictionary. Defaults to True.
            add_details (bool, optional): If True, adds task execution details to the results.
            job_data (str, optional): URL to YAML file with data or dictionary/list of data to pass on to Jinja2 rendering context.
            **kwargs: Additional arguments to pass to the task plugin.

        Returns:
            dict: A dictionary with the results of the configuration task.

        Raises:
            UnsupportedPluginError: If the specified plugin is not supported.
            FileNotFoundError: If the specified job data file cannot be downloaded.
        """
        config = config if isinstance(config, list) else [config]
        ret = Result(task=f"{self.name}:cfg", result={} if to_dict else [])

        filtered_nornir, no_match_result = self.filter_hosts_and_validate(kwargs, ret)
        if ret.status == "no_match":
            return ret

        # decide on what send commands task plugin to use
        if plugin == "netmiko":
            task_plugin = netmiko_send_config
        elif plugin == "scrapli":
            task_plugin = scrapli_send_config
        elif plugin == "napalm":
            task_plugin = napalm_configure
        else:
            raise UnsupportedPluginError(f"Plugin '{plugin}' not supported")

        job_data = self.load_job_data(job_data)

        nr = self._add_processors(filtered_nornir, kwargs, job)  # add processors

        # render config using Jinja2 on a per-host basis
        for host in nr.inventory.hosts.values():
            rendered = self.jinja2_render_templates(
                templates=config,
                context={
                    "host": host,
                    "norfab": self.client,
                    "job_data": job_data,
                    "netbox": self.add_jinja2_netbox(),
                },
                filters=self.add_jinja2_filters(),
            )
            host.data["__task__"] = {"config": rendered}

        # run task
        log.debug(
            f"{self.name} - sending config commands '{config}', kwargs '{kwargs}', is dry_run - '{dry_run}'"
        )
        if dry_run is True:
            result = nr.run(
                task=nr_test, use_task_data="config", name="dry_run", **kwargs
            )
        else:
            with self.connections_lock:
                result = nr.run(task=task_plugin, **kwargs)

        ret.failed = result.failed  # failed is true if any of the hosts failed
        ret.result = ResultSerializer(result, to_dict=to_dict, add_details=add_details)

        # remove __task__ data
        for host_name, host_object in nr.inventory.hosts.items():
            _ = host_object.data.pop("__task__", None)

        self.watchdog.connections_update(nr, plugin)
        self.watchdog.connections_clean()

        return ret

    @Task(fastapi={"methods": ["POST"]})
    def test(
        self,
        job: Job,
        suite: Union[list, str],
        subset: str = None,
        dry_run: bool = False,
        remove_tasks: bool = True,
        failed_only: bool = False,
        return_tests_suite: bool = False,
        job_data: Any = None,
        extensive: bool = False,
        groups: list = None,
        **kwargs: Any,
    ) -> Result:
        """
        Function to test networks using a suite of tests.

        Args:
            job: NorFab Job object containing relevant metadata
            suite (Union[list, str]): URL Path to YAML file with tests or a list of test definitions
                or template URL that resolves to a file path.
            subset (str, optional): List or string with comma-separated non-case-sensitive glob
                patterns to filter tests by name. Ignored if dry_run is True.
            dry_run (bool, optional): If True, returns produced per-host tests suite content only.
            remove_tasks (bool, optional): If False, results will include other tasks output.
            failed_only (bool, optional): If True, returns test results for failed tests only.
            return_tests_suite (bool, optional): If True, returns rendered per-host tests suite
                content in addition to test results using a dictionary with ``results`` and ``suite`` keys.
            job_data (str, optional): URL to YAML file with data or dictionary/list of data
                to pass on to Jinja2 rendering context.
            extensive (bool, optional): return extensive results, equivalent to using these arguments:

                - remove_tasks = False
                - return_tests_suite = True
                - add_details = True
                - to_dict = False

            groups (list, optional): list of test group names to run

            **kwargs: Any additional arguments to pass on to the Nornir service task.

        Returns:
            dict: A dictionary containing the test results. If `return_tests_suite` is True,
                the dictionary will contain both the test results and the rendered test suite.

        Note:
            Result `failed` attribute is set to True if any of the tests failed for any of the hosts.

        Raises:
            RuntimeError: If there is an error in rendering the Jinja2 templates or loading the YAML.
        """
        tests = {}  # dictionary to hold per-host test suites
        # set extensive details flags
        if extensive is True:
            kwargs["add_details"] = True
            kwargs["to_dict"] = False
            remove_tasks = False
            return_tests_suite = True
        add_details = kwargs.get("add_details", False)  # ResultSerializer
        to_dict = kwargs.get("to_dict", True)  # ResultSerializer
        filters = {k: kwargs.get(k) for k in list(kwargs.keys()) if k in FFun_functions}
        ret = Result(task=f"{self.name}:test", result={} if to_dict else [])
        suites = {}  # dictionary to hold combined test suites

        filtered_nornir, ret = self.filter_hosts_and_validate(kwargs, ret)
        if ret.status == "no_match":
            if return_tests_suite is True:
                ret.result = {"test_results": [], "suite": {}}
            return ret

        # download job data
        job_data = self.load_job_data(job_data)

        # generate per-host test suites
        for host_name, host in filtered_nornir.inventory.hosts.items():
            # render suite using Jinja2
            try:
                rendered_suite = self.jinja2_render_templates(
                    templates=[suite],
                    context={
                        "host": host,
                        "norfab": self.client,
                        "job_data": job_data,
                        "netbox": self.add_jinja2_netbox(),
                    },
                    filters=self.add_jinja2_filters(),
                )
            except Exception as e:
                msg = f"{self.name} - '{suite}' Jinja2 rendering failed: '{type(e).__name__}:{e}'"
                raise RuntimeError(msg) from e
            # load suit using YAML
            try:
                tests[host_name] = yaml.safe_load(rendered_suite) or []
            except Exception as e:
                msg = f"{self.name} - '{suite}' YAML load failed: '{type(e).__name__}:{e}'"
                raise RuntimeError(msg) from e

        # validate tests suite
        try:
            _ = modelTestsProcessorSuite(tests=tests)
        except Exception as e:
            msg = f"{self.name} - '{suite}' suite validation failed: '{type(e).__name__}:{e}'"
            raise RuntimeError(msg) from e

        # download pattern, schema and custom function files
        for host_name in tests.keys():
            for index, item in enumerate(tests[host_name]):
                for k in ["pattern", "schema", "function_file"]:
                    if self.is_url(item.get(k)):
                        item[k] = self.fetch_file(
                            item[k], raise_on_fail=True, read=True
                        )
                        if k == "function_file":
                            item["function_text"] = item.pop(k)
                tests[host_name][index] = item

        # save per-host tests suite content before mutating it
        if return_tests_suite is True:
            return_suite = copy.deepcopy(tests)

        log.debug(f"{self.name} - running test '{suite}', is dry run - '{dry_run}'")
        # run dry run task
        if dry_run is True:
            result = filtered_nornir.run(
                task=nr_test, name="tests_dry_run", ret_data_per_host=tests
            )
            ret.result = ResultSerializer(
                result, to_dict=to_dict, add_details=add_details
            )
        # combine per-host tests in suites based on task and arguments
        # Why - to run tests using any nornir service tasks with various arguments
        else:
            for host_name, host_tests in tests.items():
                for test in host_tests:
                    dhash = hashlib.md5()
                    test_args = test.pop("norfab", {})
                    nrtask = test_args.get("nrtask", "cli")
                    assert nrtask in [
                        "cli",
                        "network",
                        "cfg",
                        "task",
                    ], f"{self.name} - unsupported NorFab Nornir Service task '{nrtask}'"
                    test_json = json.dumps(test_args, sort_keys=True).encode()
                    dhash.update(test_json)
                    test_hash = dhash.hexdigest()
                    suites.setdefault(test_hash, {"params": test_args, "tests": {}})
                    suites[test_hash]["tests"].setdefault(host_name, [])
                    suites[test_hash]["tests"][host_name].append(test)
            log.debug(
                f"{self.name} - combined per-host tests suites based on NorFab Nornir Service task and arguments:\n{suites}"
            )
            # run test suites collecting output from devices
            for tests_suite in suites.values():
                nrtask = tests_suite["params"].pop("nrtask", "cli")
                function_kwargs = {
                    **tests_suite["params"],
                    **kwargs,
                    **filters,
                    "tests": tests_suite["tests"],
                    "remove_tasks": remove_tasks,
                    "failed_only": failed_only,
                    "subset": subset,
                    "groups": groups,
                }
                result = getattr(self, nrtask)(
                    job=job, **function_kwargs
                )  # returns Result object
                # save test results into overall results
                if to_dict == True:
                    for host_name, host_res in result.result.items():
                        ret.result.setdefault(host_name, {})
                        ret.result[host_name].update(host_res)
                        # set return result failed to true if any of the tests failed
                        for test_res in host_res.values():
                            if add_details:
                                if test_res["result"] != "PASS":
                                    ret.failed = True
                            elif test_res != "PASS":
                                ret.failed = True
                else:
                    ret.result.extend(result.result)
                    # set return result failed to true if any of the tests failed
                    if any(r["result"] != "PASS" for r in result.result):
                        ret.failed = True

        # check if need to return tests suite content
        if return_tests_suite is True:
            ret.result = {"test_results": ret.result, "suite": return_suite}

        return ret

    @Task(fastapi={"methods": ["POST"]})
    def network(self, job: Job, fun: str, **kwargs) -> Result:
        """
        Task to call various network-related utility functions.

        Args:
            job: NorFab Job object containing relevant metadata
            fun (str): The name of the utility function to call.
            kwargs (dict): Arguments to pass to the utility function.

        Available utility functions:

        - **resolve_dns** Resolves hosts' hostname DNS, returning IP addresses using
            `nornir_salt.plugins.tasks.network.resolve_dns` Nornir-Salt function.
        - **ping** Executes ICMP ping to host using `nornir_salt.plugins.tasks.network.ping`
            Nornir-Salt function.

        Returns:
            dict: A dictionary containing the results of the network utility function.

        Raises:
            UnsupportedPluginError: If the specified utility function is not supported.
        """
        kwargs["call"] = fun
        return self.task(
            job=job,
            plugin="nornir_salt.plugins.tasks.network",
            **kwargs,
        )

    @Task(fastapi={"methods": ["POST"]})
    def parse(
        self,
        job: Job,
        plugin: str = "napalm",
        getters: Union[str, list] = "get_facts",
        template: str = None,
        commands: Union[str, list] = None,
        to_dict: bool = True,
        add_details: bool = False,
        **kwargs: Any,
    ) -> Result:
        """
        Parse network device output using specified plugin and options.

        Args:
            job: NorFab Job object containing relevant metadata
            plugin (str): The plugin to use for parsing. Options are:

                - napalm - parse devices output using NAPALM getters
                - ttp - use TTP Templates to parse devices output
                - textfsm - use TextFSM templates to parse devices output

            getters (str): The getters to use with the "napalm" plugin.
            template (str): The template to use with the "ttp" or "textfsm" plugin.
            commands (list): The list of commands to run with the "ttp" or "textfsm" plugin.
            to_dict (bool): Whether to convert the result to a dictionary.
            add_details (bool): Whether to add details to the result.
            **kwargs: Additional keyword arguments to pass to the plugin.

        Returns:
            Result: A Result object containing the parsed data.

        Raises:
            UnsupportedPluginError: If the specified plugin is not supported.
        """
        filters = {k: kwargs.get(k) for k in list(kwargs.keys()) if k in FFun_functions}
        ret = Result(task=f"{self.name}:parse", result={} if to_dict else [])

        filtered_nornir, ret = self.filter_hosts_and_validate(kwargs, ret)
        if ret.status == "no_match":
            return ret

        if plugin == "napalm":
            nr = self._add_processors(filtered_nornir, kwargs, job)  # add processors
            result = nr.run(task=napalm_get, getters=getters, **kwargs)
            ret.result = ResultSerializer(
                result, to_dict=to_dict, add_details=add_details
            )
            ret.failed = result.failed  # failed is true if any of the hosts failed
        elif plugin == "ttp":
            result = self.cli(
                job=job,
                commands=commands or [],
                run_ttp=template,
                **filters,
                **kwargs,
                to_dict=to_dict,
                add_details=add_details,
                plugin="netmiko",
            )
            ret.result = result.result
        elif plugin == "textfsm":
            result = self.cli(
                job=job,
                commands=commands,
                **filters,
                **kwargs,
                to_dict=to_dict,
                add_details=add_details,
                use_textfsm=True,
                textfsm_template=template,
                plugin="netmiko",
            )
            ret.result = result.result
        else:
            raise UnsupportedPluginError(f"Plugin '{plugin}' not supported")

        return ret

    @Task(fastapi={"methods": ["POST"]})
    def file_copy(
        self,
        job: Job,
        source_file: str,
        plugin: str = "netmiko",
        to_dict: bool = True,
        add_details: bool = False,
        dry_run: bool = False,
        **kwargs: Any,
    ) -> Result:
        """
        Task to transfer files to and from hosts using SCP.

        Args:
            job: NorFab Job object containing relevant metadata
            source_file (str): The path or URL of the source file to be copied in
                ``nf://path/to/file`` format
            plugin (str, optional): The plugin to use for file transfer. Supported plugins:

                - netmiko - uses `netmiko_file_transfer` task plugin.

            to_dict (bool, optional): Whether to return the result as a dictionary. Defaults to True.
            add_details (bool, optional): Whether to add detailed information to the result. Defaults to False.
            dry_run (bool, optional): If True, performs a dry run without making any changes. Defaults to False.
            **kwargs: Additional arguments to pass to the file transfer plugin.

        Returns:
            dict: The result of the file copy operation.

        Raises:
            UnsupportedPluginError: If the specified plugin is not supported.
        """
        timeout = job.timeout * 0.9
        ret = Result(task=f"{self.name}:file_copy", result={} if to_dict else [])

        filtered_nornir, ret = self.filter_hosts_and_validate(kwargs, ret)
        if ret.status == "no_match":
            return ret

        # download file from broker
        if self.is_url(source_file):
            source_file_local = self.fetch_file(
                source_file, raise_on_fail=True, read=False
            )

        # decide on what send commands task plugin to use
        if plugin == "netmiko":
            task_plugin = netmiko_file_transfer
            kwargs["source_file"] = source_file_local
            kwargs.setdefault("socket_timeout", timeout / 5)
            kwargs.setdefault("dest_file", os.path.split(source_file_local)[-1])
        else:
            raise UnsupportedPluginError(f"Plugin '{plugin}' not supported")

        nr = self._add_processors(filtered_nornir, kwargs, job)  # add processors

        # run task
        log.debug(
            f"{self.name} - running file copy with arguments '{kwargs}', is dry run - '{dry_run}'"
        )
        if dry_run is True:
            result = nr.run(task=nr_test, name="file_copy_dry_run", **kwargs)
        else:
            with self.connections_lock:
                result = nr.run(task=task_plugin, **kwargs)

        ret.failed = result.failed  # failed is true if any of the hosts failed
        ret.result = ResultSerializer(result, to_dict=to_dict, add_details=add_details)

        self.watchdog.connections_update(nr, plugin)
        self.watchdog.connections_clean()

        return ret

    @Task(fastapi={"methods": ["POST"]})
    def runtime_inventory(self, job: Job, action: str, **kwargs: Any) -> Result:
        """
        Task to work with Nornir runtime (in-memory) inventory.

        Supported actions:

        - `create_host` or `create` - creates new host or replaces existing host object
        - `read_host` or `read` - read host inventory content
        - `update_host` or `update` - non recursively update host attributes if host exists
            in Nornir inventory, do not create host if it does not exist
        - `delete_host` or `delete` - deletes host object from Nornir Inventory
        - `load` - to simplify calling multiple functions
        - `read_inventory` - read inventory content for groups, default and hosts
        - `read_host_data` - to return host's data under provided path keys
        - `list_hosts` - return a list of inventory's host names
        - `list_hosts_platforms` - return a dictionary of hosts' platforms
        - `update_defaults` - non recursively update defaults attributes

        Args:
            job: NorFab Job object containing relevant metadata
            action: action to perform on inventory
            kwargs: arguments to use with the calling action
        """
        # clean up kwargs
        _ = kwargs.pop("progress", None)
        job.event(f"Performing '{action}' action")
        return Result(result=InventoryFun(self.nr, call=action, **kwargs))

    @Task(fastapi={"methods": ["POST"]})
    def netconf(
        self,
        job: Job,
        call: str,
        plugin: str = "ncclient",
        data: str = None,
        **kwargs: Any,
    ) -> Result:
        """
        Interact with devices using NETCONF protocol utilizing one of the supported plugins.

        This method provides a unified interface to interact with network devices using
        NETCONF protocol through different backend plugins.

        Args:
            job (Job): NorFab Job object containing relevant metadata.
            call (str): The ncclient manager or scrapli netconf object method to call.
            plugin (str, optional): Name of netconf plugin to use. Available plugins:

                - ``ncclient`` (default): ``nornir-salt`` built-in plugin that uses ``ncclient``
                  library to interact with devices. Uses `ncclient_call`_ task plugin.
                - ``scrapli``: Uses ``scrapli_netconf`` connection plugin that is part of
                  ``nornir_scrapli`` library. Uses `scrapli_netconf_call`_ task plugin.

            data (str, optional): Path to file for ``rpc`` method call or rpc content.
            **kwargs (Any): Additional arguments to pass to the underlying plugin.

                - method_name (str): Name of method to provide docstring for, used only by ``help`` call.

        Returns:
            Result: A Result object containing the NETCONF operation results.

        Note:
            Special ``call`` arguments/methods available:

            - ``dir``: Returns methods supported by Ncclient connection manager object.
            - ``help``: Returns ``method_name`` docstring.
            - ``transaction``: Same as ``edit_config``, but runs a more reliable workflow:

                1. Lock target configuration datastore
                2. If server supports it - Discard previous changes if any
                3. Perform configuration edit using RPC specified in ``edit_rpc`` argument
                4. If server supports it - validate configuration if ``validate`` argument is True
                5. If server supports it - do commit confirmed if ``confirmed`` argument is True
                   using ``confirm_delay`` timer with ``commit_arg`` argument
                6. If confirmed commit requested, wait for ``commit_final_delay`` timer before
                   sending final commit, final commit does not use ``commit_arg`` arguments
                7. If server supports it - do commit operation
                8. Unlock target configuration datastore
                9. If server supports it - discard all changes if any of steps 3, 4, 5 or 7 fail
                10. Return results list of dictionaries keyed by step name

        Warning:
            Beware of differences in keywords required by different plugins, e.g. ``filter`` for
            ``ncclient`` vs ``filter_``/``filters`` for ``scrapli_netconf``. Refer to modules' API
            documentation for required arguments.

        Examples:
            Examples using ``ncclient`` plugin::

                salt nrp1 nr.nc server_capabilities FB="*"
                salt nrp1 nr.nc get_config filter='["subtree", "salt://rpc/get_config_data.xml"]' source="running"
                salt nrp1 nr.nc edit_config target="running" config="salt://rpc/edit_config_data.xml" FB="ceos1"
                salt nrp1 nr.nc transaction target="candidate" config="salt://rpc/edit_config_data.xml"
                salt nrp1 nr.nc commit
                salt nrp1 nr.nc rpc data="salt://rpc/iosxe_rpc_edit_interface.xml"
                salt nrp1 nr.nc get_schema identifier="ietf-interfaces"
                salt nrp1 nr.nc get filter='<system-time xmlns="http://cisco.com/ns/yang/Cisco-IOS-XR-shellutil-oper"/>'

            Examples using ``scrapli_netconf`` plugin::

                salt nrp1 nr.nc get filter_=salt://rpc/get_config_filter_ietf_interfaces.xml plugin=scrapli
                salt nrp1 nr.nc get_config source=running plugin=scrapli
                salt nrp1 nr.nc server_capabilities FB="*" plugin=scrapli
                salt nrp1 nr.nc rpc filter_=salt://rpc/get_config_rpc_ietf_interfaces.xml plugin=scrapli
                salt nrp1 nr.nc transaction target="candidate" config="salt://rpc/edit_config_ietf_interfaces.xml" plugin=scrapli

            Python API usage from Salt-Master::

                import salt.client
                client = salt.client.LocalClient()

                task_result = client.cmd(
                    tgt="nrp1",
                    fun="nr.nc",
                    arg=["get_config"],
                    kwarg={"source": "running", "plugin": "ncclient"},
                )

            Using special calls::

                salt nrp1 nr.nc dir
                salt nrp1 nr.nc help method_name=edit_config
                salt nrp1 nr.nc transaction target="candidate" config="salt://path/to/config_file.xml" FB="*core-1"

        .. _ncclient_call: https://nornir-salt.readthedocs.io/en/latest/Tasks/ncclient_call.html
        .. _scrapli_netconf_call: https://nornir-salt.readthedocs.io/en/latest/Tasks/scrapli_netconf_call.html
        """
        # TODO implement files download and rendering
        # kwargs.setdefault(
        #     "render", ["rpc", "config", "data", "filter", "filter_", "filters"]
        # )

        # decide on plugin to use
        if plugin.lower() == "ncclient":
            task_fun = "nornir_salt.plugins.tasks.ncclient_call"
            kwargs["connection_name"] = "ncclient"
            # TODO implement filters handling

        elif plugin.lower() == "scrapli":
            task_fun = "nornir_salt.plugins.tasks.scrapli_netconf_call"
            kwargs["connection_name"] = "scrapli_netconf"
            # TODO implement filters handling

        # run task
        return self.task(job=job, plugin=task_fun, call=call, **kwargs)



import copy
import logging
import concurrent.futures

from typing import Union
from norfab.core.worker import Task, Job
from norfab.models import Result
from .netbox_models import NetboxFastApiArgs
from .netbox_exceptions import UnsupportedNetboxVersion

log = logging.getLogger(__name__)


class NetboxCircuitsTasks:

    def _map_circuit(
        self,
        job: Job,
        circuit: dict,
        ret: Result,
        instance: str,
        devices: list,
        cache: bool,
    ) -> bool:
        """
        ThreadPoolExecutor target function to retrieve circuit details from Netbox

        Args:
            circuit (dict): The circuit data to be mapped.
            ret (Result): The result object to store the mapped data.
            instance (str): The instance of the Netbox API to use.
            devices (list): List of devices to check against the circuit endpoints.
            cache (bool): Flag to determine if the data should be cached.

        Returns:
            bool: True if the mapping is successful, False otherwise.
        """
        cid = circuit.pop("cid")
        ckt_cache_data = {}  # ckt data dictionary to save in cache
        circuit["tags"] = [i["name"] for i in circuit["tags"]]
        circuit["type"] = circuit["type"]["name"]
        circuit["provider"] = circuit["provider"]["name"]
        circuit["tenant"] = circuit["tenant"]["name"] if circuit["tenant"] else None
        circuit["provider_account"] = (
            circuit["provider_account"]["name"] if circuit["provider_account"] else None
        )
        termination_a = circuit["termination_a"]
        termination_z = circuit["termination_z"]
        termination_a = termination_a["id"] if termination_a else None
        termination_z = termination_z["id"] if termination_z else None

        msg = f"{cid} tracing circuit terminations path"
        log.info(msg)
        job.event(msg)

        # retrieve A or Z termination path using Netbox REST API
        circuit_path = None
        if termination_a is not None:
            resp = self.rest(
                job=job,
                instance=instance,
                method="get",
                api=f"circuits/circuit-terminations/{termination_a}/paths",
            )
            circuit_path = resp.result
        elif termination_z is not None:
            resp = self.rest(
                job=job,
                instance=instance,
                method="get",
                api=f"circuits/circuit-terminations/{termination_z}/paths",
            )
            circuit_path = resp.result

        # check if circuit ends connect to device or provider network
        if (
            not circuit_path
            or "name" not in circuit_path[0]["path"][0][0]
            or "name" not in circuit_path[0]["path"][-1][-1]
        ):
            msg = f"{cid} does not have two terminations, cannot trace the path"
            log.warning(msg)
            job.event(msg)
            return True

        # form A and Z connection endpoints
        end_a = {
            "device": circuit_path[0]["path"][0][0]
            .get("device", {})
            .get("name", False),
            "provider_network": "provider-network"
            in circuit_path[0]["path"][0][0]["url"],
            "name": circuit_path[0]["path"][0][0]["name"],
        }
        end_z = {
            "device": circuit_path[0]["path"][-1][-1]
            .get("device", {})
            .get("name", False),
            "provider_network": "provider-network"
            in circuit_path[0]["path"][-1][-1]["url"],
            "name": circuit_path[0]["path"][-1][-1]["name"],
        }
        circuit["is_active"] = circuit_path[0]["is_active"]

        # map path ends to devices
        if not end_a["device"] and not end_z["device"]:
            msg = f"{cid} path trace ends have no devices connected"
            log.error(msg)
            job.event(msg, severity="ERROR")
            return True
        if end_a["device"]:
            device_data = copy.deepcopy(circuit)
            device_data["interface"] = end_a["name"]
            if end_z["device"]:
                device_data["remote_device"] = end_z["device"]
                device_data["remote_interface"] = end_z["name"]
            elif end_z["provider_network"]:
                device_data["provider_network"] = end_z["name"]
            # save device data in cache
            ckt_cache_data[end_a["device"]] = device_data
            # include device data in result
            if end_a["device"] in devices:
                ret.result[end_a["device"]][cid] = device_data
        if end_z["device"]:
            device_data = copy.deepcopy(circuit)
            device_data["interface"] = end_z["name"]
            if end_a["device"]:
                device_data["remote_device"] = end_a["device"]
                device_data["remote_interface"] = end_a["name"]
            elif end_a["provider_network"]:
                device_data["provider_network"] = end_a["name"]
            # save device data in cache
            ckt_cache_data[end_z["device"]] = device_data
            # include device data in result
            if end_z["device"] in devices:
                ret.result[end_z["device"]][cid] = device_data

        # save data to cache
        if cache != False:
            ckt_cache_key = f"get_circuits::{cid}"
            if ckt_cache_data:
                self.cache.set(ckt_cache_key, ckt_cache_data, expire=self.cache_ttl)
                log.info(
                    f"{self.name}:get_circuits - {cid} cached circuit data for future use"
                )

        msg = f"{cid} circuit data mapped to devices using data from Netbox"
        log.info(msg)
        job.event(msg)
        return True

    @Task(fastapi={"methods": ["GET"], "schema": NetboxFastApiArgs.model_json_schema()})
    def get_circuits(
        self,
        job: Job,
        devices: list,
        cid: Union[None, list] = None,
        instance: Union[None, str] = None,
        dry_run: bool = False,
        cache: Union[None, bool, str] = None,
        add_interface_details: bool = False,
    ) -> Result:
        """
        Retrieve circuit information for specified devices from Netbox.

        Args:
            job: NorFab Job object containing relevant metadata
            devices (list): List of device names to retrieve circuits for.
            cid (list, optional): List of circuit IDs to filter by.
            instance (str, optional): Netbox instance to query.
            dry_run (bool, optional): If True, perform a dry run without making changes. Defaults to False.
            add_interface_details (bool, optional): If True, add interface details using `get_interfaces` call
                including interface subinterfaces - ip addresses, vrf, child interfaces with their IPs and vrf
            cache (Union[bool, str], optional): Cache usage options:

                - True: Use data stored in cache if it is up to date, refresh it otherwise.
                - False: Do not use cache and do not update cache.
                - "refresh": Ignore data in cache and replace it with data fetched from Netbox.
                - "force": Use data in cache without checking if it is up to date.

        Returns:
            dict: dictionary keyed by device names with circuits data:

                ```
                nf#netbox get circuits device-list fceos5
                {
                    "netbox-worker-1.1": {
                        "fceos5": {
                            "CID1": {
                                "comments": "",
                                "commit_rate": null,
                                "custom_fields": {},
                                "description": "",
                                "interface": "eth101",
                                "is_active": true,
                                "provider": "Provider1",
                                "provider_account": "",
                                "remote_device": "fceos4",
                                "remote_interface": "eth101",
                                "status": "active",
                                "tags": [],
                                "tenant": null,
                                "type": "DarkFibre"
                            }
                        }
                    }
                }
                nf#
                ```

        If `add_interface_details` is True returns this extra information:

        ```
        {
            "netbox-worker-1.1": {
                "fceos4": {
                    "CID2": {
                        "child_interfaces": [
                            {
                                "ip_addresses": [
                                    {
                                        "address": "10.0.0.12/24",
                                        ...
                                    }
                                ],
                                "name": "eth11.123",
                                "vrf": {
                                    "name": "MGMT"
                                }
                            }
                        ],
                        "interface": "eth11",
                        "ip_addresses": [
                            {
                                "address": "10.0.0.14/24",
                                ...
                            }
                        ],
                        "vrf": {
                            "name": "OOB_CTRL"
                        }
                    }
                }
            }
        }
        ```
        """
        cid = cid or []
        log.info(
            f"{self.name}:get_circuits - {instance or self.default_instance} Netbox, "
            f"devices {', '.join(devices)}, cid {cid}"
        )
        instance = instance or self.default_instance

        # form final result object
        ret = Result(
            task=f"{self.name}:get_circuits",
            result={d: {} for d in devices},
            resources=[instance],
        )
        cache = self.cache_use if cache is None else cache
        cid = cid or []
        circuit_fields = [
            "cid",
            "tags {name}",
            "provider {name}",
            "commit_rate",
            "description",
            "status",
            "type {name}",
            "provider_account {name}",
            "tenant {name}",
            "termination_a {id last_updated}",
            "termination_z {id last_updated}",
            "custom_fields",
            "comments",
            "last_updated",
        ]

        # form initial circuits filters based on devices' sites and cid list
        circuits_filters = {}
        device_data = self.get_devices(
            job=job, devices=copy.deepcopy(devices), instance=instance, cache=cache
        )
        sites = list(set([i["site"]["slug"] for i in device_data.result.values()]))
        if self.nb_version[instance] >= (4, 4, 0):
            slist = str(sites).replace("'", '"')  # swap quotes
            if cid:
                clist = str(cid).replace("'", '"')  # swap quotes
                circuits_filters = "{terminations: {site: {slug: {in_list: slist}}}, cid: {in_list: clist}}"
                circuits_filters = circuits_filters.replace("slist", slist).replace(
                    "clist", clist
                )
            else:
                circuits_filters = "{terminations: {site: {slug: {in_list: slist }}}}"
                circuits_filters = circuits_filters.replace("slist", slist)
        else:
            raise UnsupportedNetboxVersion(
                f"{self.name} - Netbox version {self.nb_version[instance]} is not supported, "
                f"minimum required version is {self.compatible_ge_v4}"
            )

        log.info(
            f"{self.name}:get_circuits - constructed circuits filters: '{circuits_filters}'"
        )

        if cache == True or cache == "force":
            log.info(f"{self.name}:get_circuits - retrieving circuits data from cache")
            job.event("retrieving circuits data from cache")
            cid_list = []  #  new cid list for follow up query
            # retrieve last updated data from Netbox for circuits and their terminations
            last_updated = self.graphql(
                job=job,
                obj="circuit_list",
                filters=circuits_filters,
                fields=[
                    "cid",
                    "last_updated",
                    "termination_a {id last_updated}",
                    "termination_z {id last_updated}",
                ],
                dry_run=dry_run,
                instance=instance,
            )
            last_updated.raise_for_status(f"{self.name} - get circuits query failed")

            # return dry run result
            if dry_run:
                ret.result["get_circuits_dry_run"] = last_updated.result
                return ret

            # retrieve circuits data from cache
            self.cache.expire()  # remove expired items from cache
            for device in devices:
                for circuit in last_updated.result:
                    circuit_cache_key = f"get_circuits::{circuit['cid']}"
                    log.info(
                        f"{self.name}:get_circuits - searching cache for key {circuit_cache_key}"
                    )
                    # check if cache is up to date and use it if so
                    if circuit_cache_key in self.cache:
                        cache_ckt = self.cache[circuit_cache_key]
                        # check if device uses this circuit
                        if device not in cache_ckt:
                            continue
                        # use cache forcefully
                        if cache == "force":
                            ret.result[device][circuit["cid"]] = cache_ckt[device]
                        # check circuit cache is up to date
                        if cache_ckt[device]["last_updated"] != circuit["last_updated"]:
                            continue
                        if (
                            cache_ckt[device]["termination_a"]
                            and circuit["termination_a"]
                            and cache_ckt[device]["termination_a"]["last_updated"]
                            != circuit["termination_a"]["last_updated"]
                        ):
                            continue
                        if (
                            cache_ckt[device]["termination_z"]
                            and circuit["termination_z"]
                            and cache_ckt[device]["termination_z"]["last_updated"]
                            != circuit["termination_z"]["last_updated"]
                        ):
                            continue
                        ret.result[device][circuit["cid"]] = cache_ckt[device]
                        log.info(
                            f"{self.name}:get_circuits - {circuit['cid']} retrieved data from cache"
                        )
                    elif circuit["cid"] not in cid_list:
                        cid_list.append(circuit["cid"])
                        log.info(
                            f"{self.name}:get_circuits - {circuit['cid']} no cache data found, fetching from Netbox"
                        )
            # form new filters dictionary to fetch remaining circuits data
            circuits_filters = {}
            if cid_list:
                cid_list = str(cid_list).replace("'", '"')  # swap quotes
                if self.nb_version[instance] >= (4, 4, 0):
                    circuits_filters = "{cid: {in_list: cid_list}}"
                    circuits_filters = circuits_filters.replace("cid_list", cid_list)
                else:
                    raise UnsupportedNetboxVersion(
                        f"{self.name} - Netbox version {self.nb_version[instance]} is not supported, "
                        f"minimum required version is {self.compatible_ge_v4}"
                    )
        # ignore cache data, fetch circuits from netbox
        elif cache == False or cache == "refresh":
            pass

        if circuits_filters:
            job.event("fetching circuits data from Netbox")
            query_result = self.graphql(
                job=job,
                obj="circuit_list",
                filters=circuits_filters,
                fields=circuit_fields,
                dry_run=dry_run,
                instance=instance,
            )
            query_result.raise_for_status(f"{self.name} - get circuits query failed")

            # return dry run result
            if dry_run is True:
                return query_result

            all_circuits = query_result.result

            # iterate over circuits and map them to devices
            msg = (
                f"retrieved data for {len(all_circuits)} "
                f"circuits from Netbox, mapping circuits to devices"
            )
            log.info(msg)
            job.event(msg)
            with concurrent.futures.ThreadPoolExecutor(max_workers=10) as executor:
                results = [
                    executor.submit(
                        self._map_circuit, job, circuit, ret, instance, devices, cache
                    )
                    for circuit in all_circuits
                ]
                for _ in concurrent.futures.as_completed(results):
                    continue

        if add_interface_details:
            job.event("fetching circuits interface details")
            # collect devices and interfaces to get details for
            fetch_interfaces = set()
            fetch_devices = set(ret.result.keys())
            for device_name, circuits in ret.result.items():
                for ckt_data in circuits.values():
                    fetch_interfaces.add(ckt_data["interface"])
                    if ckt_data.get("remote_device"):
                        fetch_devices.add(ckt_data["remote_device"])
                        fetch_interfaces.add(ckt_data["remote_interface"])
            # fetch interfaces data with IP addresses
            interfaces_data = self.get_interfaces(
                job=job,
                devices=list(fetch_devices),
                interface_list=list(fetch_interfaces),
                ip_addresses=True,
            ).result
            # map interfaces details to circuits
            for device_name, circuits in ret.result.items():
                for circuit_id, ckt_data in circuits.items():
                    interface_name = ckt_data["interface"]
                    if interface_name in interfaces_data.get(device_name, {}):
                        interface_data = interfaces_data[device_name][interface_name]
                        ckt_data["child_interfaces"] = interface_data.get(
                            "child_interfaces", []
                        )
                        ckt_data["ip_addresses"] = interface_data.get(
                            "ip_addresses", []
                        )
                        ckt_data["vrf"] = interface_data.get("vrf", {})
                    else:
                        log.error(
                            f"{device_name}:{circuit_id} failed to find '{interface_name}' interface details"
                        )

        return ret



import json
import builtins

from enum import Enum
from pydantic import (
    BaseModel,
    StrictBool,
    StrictInt,
    StrictStr,
    Field,
)
from ..common import ClientRunJobArgs, log_error_or_result, listen_events
from typing import Union, Optional, List, Dict
from picle.models import Outputters, PipeFunctionsModel
from .nornir_picle_shell_common import NorniHostsFilters


class CreateHostModel(ClientRunJobArgs):
    workers: Union[StrictStr, List[StrictStr]] = Field(
        "any", description="Nornir workers to target"
    )
    name: StrictStr = Field(..., description="Name of the host")
    username: StrictInt = Field(None, description="Host connections username")
    password: StrictInt = Field(None, description="Host connections password")
    platform: StrictInt = Field(
        None, description="Host platform recognized by connection plugin"
    )
    hostname: StrictStr = Field(
        None,
        description="Hostname of the host to initiate connection with, IP address or FQDN",
    )
    port: StrictInt = Field(22, description="TCP port to initiate connection with")
    connection_options: Dict = Field(
        None,
        description="JSON string with connection options",
        alias="connection-options",
    )
    groups: List[StrictStr] = Field(
        None, description="List of groups to associate with this host"
    )
    data: Dict = Field(None, description="JSON string with arbitrary host data")
    progress: Optional[StrictBool] = Field(
        True,
        description="Display progress events",
        json_schema_extra={"presence": True},
    )

    @staticmethod
    @listen_events
    def run(uuid, *args, **kwargs):
        NFCLIENT = builtins.NFCLIENT
        workers = kwargs.pop("workers")
        timeout = kwargs.pop("timeout", 600)
        kwargs["action"] = "create_host"
        verbose_result = kwargs.pop("verbose_result", False)
        nowait = kwargs.pop("nowait", False)

        if kwargs.get("connection_options"):
            kwargs["connection_options"] = json.loads(kwargs["connection_options"])
        if kwargs.get("data"):
            kwargs["data"] = json.loads(kwargs["data"])
        if kwargs.get("groups") and isinstance(kwargs["groups"], str):
            kwargs["groups"] = [kwargs["groups"]]

        result = NFCLIENT.run_job(
            "nornir",
            "runtime_inventory",
            workers=workers,
            args=args,
            kwargs=kwargs,
            uuid=uuid,
            timeout=timeout,
            nowait=nowait,
        )

        if nowait:
            return result, Outputters.outputter_nested

        return log_error_or_result(result, verbose_result=verbose_result)

    class PicleConfig:
        outputter = Outputters.outputter_nested


class GroupsUpdateAction(str, Enum):
    append = "append"
    insert = "insert"
    remove = "remove"


class UpdateHostModel(ClientRunJobArgs):
    workers: Union[StrictStr, List[StrictStr]] = Field(
        "all", description="Nornir workers to target"
    )
    name: StrictStr = Field(..., description="Name of the host")
    username: StrictInt = Field(None, description="Host connections username")
    password: StrictInt = Field(None, description="Host connections password")
    platform: StrictInt = Field(
        None, description="Host platform recognized by connection plugin"
    )
    hostname: StrictStr = Field(
        None,
        description="Hostname of the host to initiate connection with, IP address or FQDN",
    )
    port: StrictInt = Field(22, description="TCP port to initiate connection with")
    connection_options: Dict = Field(
        None,
        description="JSON string with connection options",
        alias="connection-options",
    )
    groups: List[StrictStr] = Field(
        None, description="List of groups to associate with this host"
    )
    groups_action: GroupsUpdateAction = Field(
        "append", description="Action to perform with groups", alias="groups-action"
    )
    data: Dict = Field(None, description="JSON string with arbitrary host data")
    progress: Optional[StrictBool] = Field(
        True,
        description="Display progress events",
        json_schema_extra={"presence": True},
    )

    class PicleConfig:
        pipe = PipeFunctionsModel
        outputter = Outputters.outputter_nested

    @staticmethod
    @listen_events
    def run(uuid, *args, **kwargs):
        NFCLIENT = builtins.NFCLIENT
        workers = kwargs.pop("workers")
        timeout = kwargs.pop("timeout", 600)
        kwargs["action"] = "update_host"
        verbose_result = kwargs.pop("verbose_result", False)
        nowait = kwargs.pop("nowait", False)

        if kwargs.get("connection_options"):
            kwargs["connection_options"] = json.loads(kwargs["connection_options"])
        if kwargs.get("data"):
            kwargs["data"] = json.loads(kwargs["data"])
        if kwargs.get("groups") and isinstance(kwargs["groups"], str):
            kwargs["groups"] = [kwargs["groups"]]

        result = NFCLIENT.run_job(
            "nornir",
            "runtime_inventory",
            workers=workers,
            args=args,
            kwargs=kwargs,
            uuid=uuid,
            timeout=timeout,
            nowait=nowait,
        )

        if nowait:
            return result, Outputters.outputter_nested

        return log_error_or_result(result, verbose_result=verbose_result)


class DeleteHostModel(ClientRunJobArgs):
    workers: Union[StrictStr, List[StrictStr]] = Field(
        "all", description="Nornir workers to target"
    )
    name: StrictStr = Field(..., description="Name of the host")
    progress: Optional[StrictBool] = Field(
        True,
        description="Display progress events",
        json_schema_extra={"presence": True},
    )

    @staticmethod
    @listen_events
    def run(uuid, *args, **kwargs):
        NFCLIENT = builtins.NFCLIENT
        workers = kwargs.pop("workers")
        timeout = kwargs.pop("timeout", 600)
        kwargs["action"] = "delete_host"
        verbose_result = kwargs.pop("verbose_result", False)
        nowait = kwargs.pop("nowait", False)

        result = NFCLIENT.run_job(
            "nornir",
            "runtime_inventory",
            workers=workers,
            args=args,
            kwargs=kwargs,
            uuid=uuid,
            timeout=timeout,
            nowait=nowait,
        )

        if nowait:
            return result, Outputters.outputter_nested

        return log_error_or_result(result, verbose_result=verbose_result)

    class PicleConfig:
        outputter = Outputters.outputter_nested


class ReadHostDataKeyModel(NorniHostsFilters, ClientRunJobArgs):
    workers: Union[StrictStr, List[StrictStr]] = Field(
        "all", description="Nornir workers to target"
    )
    keys: Union[StrictStr, List[StrictStr]] = Field(
        ...,
        description="Dot separated path within host data",
        examples="config.interfaces.Lo0",
    )
    progress: Optional[StrictBool] = Field(
        True,
        description="Display progress events",
        json_schema_extra={"presence": True},
    )

    @staticmethod
    @listen_events
    def run(uuid, *args, **kwargs):
        NFCLIENT = builtins.NFCLIENT
        workers = kwargs.pop("workers")
        timeout = kwargs.pop("timeout", 600)
        kwargs["action"] = "read_host_data"
        verbose_result = kwargs.pop("verbose_result", False)
        nowait = kwargs.pop("nowait", False)

        if isinstance(kwargs["keys"], str):
            kwargs["keys"] = [kwargs["keys"]]

        result = NFCLIENT.run_job(
            "nornir",
            "runtime_inventory",
            workers=workers,
            args=args,
            kwargs=kwargs,
            uuid=uuid,
            timeout=timeout,
            nowait=nowait,
        )

        if nowait:
            return result, Outputters.outputter_nested

        return log_error_or_result(result, verbose_result=verbose_result)

    class PicleConfig:
        outputter = Outputters.outputter_json


class InventoryLoadContainerlabModel(ClientRunJobArgs):
    workers: Union[StrictStr, List[StrictStr]] = Field(
        ...,
        description="Nornir workers to load inventory into",
    )
    clab_workers: Union[StrictStr, List[StrictStr]] = Field(
        None,
        description="Containerlab workers to load inventory from",
        alias="clab-workers",
    )
    progress: Optional[StrictBool] = Field(
        True,
        description="Display progress events",
        json_schema_extra={"presence": True},
    )
    lab_name: StrictStr = Field(
        None,
        description="Name of Containerlab lab to load hosts' inventory",
        alias="lab-name",
    )
    groups: Union[StrictStr, List[StrictStr]] = Field(
        None,
        description="List of Nornir groups to associate with hosts",
    )
    use_default_credentials: StrictBool = Field(
        None,
        description="Use Containerlab default credentials for all hosts",
        alias="use-default-credentials",
    )
    dry_run: StrictBool = Field(
        None,
        description="Do not refresh Nornir, only return pulled inventory",
        json_schema_extra={"presence": True},
        alias="dry-run",
    )

    class PicleConfig:
        outputter = Outputters.outputter_nested
        pipe = PipeFunctionsModel

    @staticmethod
    def source_workers():
        NFCLIENT = builtins.NFCLIENT
        reply = NFCLIENT.mmi(
            "mmi.service.broker", "show_workers", kwargs={"service": "nornir"}
        )
        workers = [i["name"] for i in reply["results"]]

        return ["all", "any"] + workers

    @staticmethod
    def source_lab_name():
        NFCLIENT = builtins.NFCLIENT
        ret = []
        result = NFCLIENT.run_job("containerlab", "get_running_labs")
        for wname, wres in result.items():
            ret.extend(wres["result"])
        return ret

    @staticmethod
    def source_clab_workers():
        NFCLIENT = builtins.NFCLIENT
        reply = NFCLIENT.mmi(
            "mmi.service.broker", "show_workers", kwargs={"service": "containerlab"}
        )
        workers = [i["name"] for i in reply["results"]]

        return ["all", "any"] + workers

    @staticmethod
    @listen_events
    def run(uuid, **kwargs):
        NFCLIENT = builtins.NFCLIENT
        workers = kwargs.pop("workers")
        timeout = kwargs.pop("timeout", 600)
        verbose_result = kwargs.pop("verbose_result")
        nowait = kwargs.pop("nowait", False)

        if isinstance(kwargs.get("groups"), str):
            kwargs["groups"] = [kwargs["groups"]]

        result = NFCLIENT.run_job(
            "nornir",
            "nornir_inventory_load_containerlab",
            kwargs=kwargs,
            workers=workers,
            timeout=timeout,
            uuid=uuid,
            nowait=nowait,
        )

        if nowait:
            return result, Outputters.outputter_nested

        return log_error_or_result(result, verbose_result=verbose_result)


class InventoryLoadModel(BaseModel):
    containerlab: InventoryLoadContainerlabModel = Field(
        None, description="Load inventory from running Containerlab lab(s)"
    )


class NornirInventoryShell(BaseModel):
    create_host: CreateHostModel = Field(
        None, description="Create new host", alias="create-host"
    )
    update_host: UpdateHostModel = Field(
        None, description="Update existing host details", alias="update-host"
    )
    delete_host: DeleteHostModel = Field(
        None, description="Delete host from inventory", alias="delete-host"
    )
    read_host_data: ReadHostDataKeyModel = Field(
        None,
        description="Return host data at given dor-separated key path",
        alias="read-host-data",
    )
    load: InventoryLoadModel = Field(
        None, description="Load inventory from external source"
    )

    class PicleConfig:
        subshell = True
        prompt = "nf[nornir-inventory]#"



import logging
import sys
import time
import os
import signal
import importlib.metadata
import subprocess
import yaml
import json
import socket
import ipaddress

from norfab.core.worker import NFPWorker, Task, Job
from norfab.core.inventory import merge_recursively
from norfab.models.containerlab import DeployTask, DeployTaskResponse
from norfab.models import Result
from norfab.utils.platform_map import PlatformMap
from typing import Union, Tuple

SERVICE = "containerlab"

log = logging.getLogger(__name__)


class ContainerlabWorker(NFPWorker):
    """
    FastAPContainerlabWorker IWorker is a worker class that integrates with containerlab to run network topologies.

    Args:
        inventory (str): Inventory configuration for the worker.
        broker (str): Broker URL to connect to.
        worker_name (str): Name of this worker.
        exit_event (threading.Event, optional): Event to signal worker to stop/exit.
        init_done_event (threading.Event, optional): Event to signal when worker is done initializing.
        log_level (str, optional): Logging level for this worker.
        log_queue (object, optional): Queue for logging.
    """

    def __init__(
        self,
        inventory: str,
        broker: str,
        worker_name: str,
        exit_event=None,
        init_done_event=None,
        log_level: str = None,
        log_queue: object = None,
    ):
        super().__init__(
            inventory, broker, SERVICE, worker_name, exit_event, log_level, log_queue
        )
        self.init_done_event = init_done_event
        self.exit_event = exit_event

        # create directory to store lab topologies
        self.topologies_dir = os.path.join(self.base_dir, "topologies")
        os.makedirs(self.topologies_dir, exist_ok=True)

        # merge local inventory with inventory from broker
        merge_recursively(self.inventory[self.name], self.load_inventory())

        self.clab_version = self.get_clab_version()

        self.init_done_event.set()

    def get_clab_version(self):
        clab_version = None
        clab_version = subprocess.run(
            ["containerlab", "version"], capture_output=True, text=True
        )
        if clab_version.returncode == 0:
            clab_version = clab_version.stdout
            for line in clab_version.splitlines()[6:]:
                if "version" in line.lower():
                    clab_version = [int(i) for i in line.split(" ")[-1].split(".")]
                    break
            else:
                raise RuntimeError(
                    "Containerlab worker failed to get containerlab version"
                )
        else:
            raise RuntimeError(
                f"Containerlab worker failed to get containerlab version, "
                f"error: '{clab_version.stderr.decode('utf-8')}'"
            )

        return clab_version

    def worker_exit(self):
        """
        Terminates the current process by sending a SIGTERM signal to itself.

        This method retrieves the current process ID using `os.getpid()` and then
        sends a SIGTERM signal to terminate the process using `os.kill()`.
        """
        os.kill(os.getpid(), signal.SIGTERM)

    @Task(fastapi={"methods": ["GET"]})
    def get_version(self) -> Result:
        """
        Produce a report of the versions of various Python packages.

        This method collects the versions of several specified Python packages
        and returns them in a dictionary.

        Returns:
            Result: An object containing the task name and a dictionary with
                    the package names as keys and their respective versions as values.
        """
        libs = {
            "norfab": "",
            "pydantic": "",
            "python": sys.version.split(" ")[0],
            "platform": sys.platform,
            "containerlab": ".".join([str(i) for i in self.clab_version]),
        }
        ret = Result(task=f"{self.name}:get_version", result=libs)

        # get version of packages installed
        for pkg in libs.keys():
            try:
                libs[pkg] = importlib.metadata.version(pkg)
            except importlib.metadata.PackageNotFoundError:
                pass

        return ret

    @Task(fastapi={"methods": ["GET"]})
    def get_inventory(self) -> Result:
        """
        Retrieve the inventory of the Containerlab worker.

        Returns:
            Dict: A dictionary containing the combined inventory of Containerlab.
        """
        return Result(
            result=self.inventory[self.name],
            task=f"{self.name}:get_inventory",
        )

    @Task(fastapi={"methods": ["GET"]})
    def get_containerlab_status(self) -> Result:
        """
        Retrieve the status of the Containerlab worker.

        Returns:
            Result: A result object containing the status of the Containerlab worker.
        """
        status = "OS NOT SUPPORTED" if sys.platform.startswith("win") else "READY"
        return Result(
            task=f"{self.name}:get_containerlab_status",
            result={"status": status},
        )

    @Task(fastapi={"methods": ["GET"]})
    def get_running_labs(self, job: Job, timeout: int = None) -> Result:
        """
        Retrieve a list of running containerlab lab names.

        This method inspects the current state of containerlab and returns
        the names of labs that are currently running. The names are sorted
        and duplicates are removed.

        Args:
            timeout (int, optional): The timeout value in seconds for the inspection
                operation. Defaults to None.

        Returns:
            Result: A Result object containing the task name and a list of running
            lab names.
        """
        timeout = timeout or 600
        ret = Result(task=f"{self.name}:get_running_labs", result=[])
        inspect = self.inspect(job=job, timeout=timeout)

        # form topologies list if any of them are running
        if inspect.result:
            ret.result = inspect.result.keys()
            ret.result = list(sorted(set(ret.result)))

        return ret

    @Task(fastapi={"methods": ["POST"]})
    def run_containerlab_command(
        self,
        job: Job,
        args: list,
        cwd: str = None,
        timeout: int = None,
        ret: Result = None,
        env: Union[None, dict] = None,
        expect_output: bool = True,
    ) -> Tuple:
        """
        Executes a containerlab command using subprocess and processes its output.

        Args:
            args (list): The list of command-line arguments to execute.
            cwd (str, optional): The working directory to execute the command in. Defaults to None.
            timeout (int, optional): The timeout for the command execution in seconds. Defaults to None.
            ret (Result, optional): An optional Norfab result object to populate with the command's output. Defaults to None.
            env (dict, Optional): OS Environment variables to use when running the process
            expect_output (bool, Optional): whether to expect any output from command

        Returns:
            Tuple: If `ret` is None, returns a tuple containing:
                - output (str): The standard output of the command.
                - logs (list): A list of log messages from the command's standard error.
                - proc (subprocess.Popen): The subprocess object for the executed command.
            Result: If `ret` is provided, returns the populated `Result` object with the following attributes:
                - result: The parsed JSON output or raw output of the command.
                - failed (bool): Indicates if the command execution failed.
                - errors (list): A list of error messages if the command failed.
                - messages (list): A list of log messages if the command succeeded.

        Raises:
            Exception: If the output cannot be parsed as JSON when `ret` is provided.

        Notes:
            - The method reads the command's standard error line by line and processes messages containing "msg=".
            - If the command fails (non-zero return code), the `ret.failed` attribute is set to True, and errors are populated.
            - If the command succeeds, the `ret.messages` attribute is populated with log messages.
        """
        timeout = timeout or 600
        output, logs = "", []
        begin = time.time()
        timeout = timeout or 600
        env = env or dict(os.environ)

        with subprocess.Popen(
            args,
            cwd=cwd,
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
            text=True,
            env=env,
        ) as proc:
            while proc.poll() is None:
                if time.time() - begin > timeout:
                    raise TimeoutError(
                        f"Containerlab output collection {timeout}s timeout expired."
                    )
                msg = proc.stderr.readline().strip()
                if msg:
                    job.event(msg.split("msg=")[-1].replace('\\"', "").strip('"'))
                    logs.append(msg)
                time.sleep(0.01)
            # read remaining messages
            for msg in proc.stderr.readlines():
                msg = msg.strip()
                if msg:
                    job.event(msg.split("msg=")[-1].replace('\\"', "").strip('"'))
                    logs.append(msg)
                time.sleep(0.01)
            # read process output
            output = proc.stdout.read()

        # populate Norfab result object
        if ret is not None:
            # check if command failed
            if proc.returncode != 0:
                ret.failed = True
                ret.errors = ["\n".join(logs)]
            # check if got no output
            elif not output.strip() and expect_output is True:
                ret.failed = True
                ret.errors = ["\n".join(logs)]
            else:
                ret.messages = ["\n".join(logs)]
                try:
                    ret.result = json.loads(output)
                except Exception:
                    # if failed, remove any beginning lines that are not part of json
                    try:
                        line_split = output.splitlines()
                        for index, line in enumerate(line_split):
                            # find first json output line
                            if "{" in line or "[" in line:
                                ret.result = json.loads("\n".join(line_split[index:]))
                                break
                    except Exception as e:
                        ret.result = output
                        log.error(
                            f"{self.name} - failed to load containerlab results into JSON, error: {e}, result: '{output}'"
                        )

            return ret
        # return command results as is
        else:
            return output, logs, proc

    @Task(fastapi={"methods": ["POST"]}, input=DeployTask, output=DeployTaskResponse)
    def deploy(
        self,
        job: Job,
        topology: str,
        reconfigure: bool = False,
        timeout: int = None,
        node_filter: Union[None, str] = None,
    ) -> Result:
        """
        Deploys a containerlab topology.

        This method handles the deployment of a containerlab topology by downloading
        the topology file, organizing it into a specific folder structure, and executing
        the `containerlab deploy` command with the appropriate arguments.

        Args:
            topology (str): The path to the topology file to be deployed.
            reconfigure (bool, optional): If True, reconfigures an already deployed lab.
                Defaults to False.
            timeout (int, optional): The timeout in seconds for the deployment process.
                Defaults to None (no timeout).
            node_filter (str, optional): A filter to specify which nodes to deploy.

        Returns:
            Result: deployment results with a list of nodes deployed

        Raises:
            Exception: If the topology file cannot be fetched.
        """
        timeout = timeout or 600
        ret = Result(task=f"{self.name}:deploy")

        # create folder to store topology
        topology_folder = os.path.split(os.path.split(topology)[0])[-1]
        topology_folder = os.path.join(self.topologies_dir, topology_folder)
        os.makedirs(topology_folder, exist_ok=True)

        # download topology file
        topology_file = os.path.join(topology_folder, os.path.split(topology)[-1])
        if self.is_url(topology):
            downloaded_topology_file = self.fetch_file(
                topology, raise_on_fail=True, read=False
            )
            os.rename(
                downloaded_topology_file, topology_file
            )  # move topology file under desired folder

        # form command arguments
        args = ["containerlab", "deploy", "-f", "json", "-t", topology_file]
        if reconfigure is True:
            args.append("--reconfigure")
            job.event(f"Re-deploying lab {os.path.split(topology_file)[-1]}")
        else:
            job.event(f"Deploying lab {os.path.split(topology_file)[-1]}")
        if node_filter:
            args.append("--node-filter")
            args.append(node_filter)

        # add needed env variables
        env = dict(os.environ)
        env["CLAB_VERSION_CHECK"] = "disable"

        # run containerlab command
        return self.run_containerlab_command(
            args=args, cwd=topology_folder, timeout=timeout, ret=ret, env=env, job=job
        )

    @Task(fastapi={"methods": ["DELETE"]})
    def destroy_lab(self, lab_name: str, job: Job, timeout: int = None) -> Result:
        """
        Destroys a specified lab.

        Args:
            lab_name (str): The name of the lab to be destroyed.
            timeout (int, optional): The timeout value in seconds for the operation. Defaults to None.

        Returns:
            Result: An object containing the status of the operation, errors (if any),
                    and the result indicating whether the lab was successfully destroyed.

        Behavior:
            - Retrieves the lab details using the `inspect` method.
            - If the lab is not found, marks the operation as failed and returns an error.
            - If the lab is found, retrieves the topology file and its folder.
            - Executes the `containerlab destroy` command using the topology file.
            - Updates the result to indicate success or failure of the destruction process.
        """
        timeout = timeout or 600
        ret = Result(task=f"{self.name}:destroy_lab")

        # get lab details
        inspect = self.inspect(
            job=job, timeout=timeout, lab_name=lab_name, details=True
        )

        if not inspect.result:
            ret.failed = True
            ret.errors = [f"'{lab_name}' lab not found"]
            ret.result = {lab_name: False}
        else:
            topology_file = inspect.result[lab_name][0]["Labels"]["clab-topo-file"]
            topology_folder = os.path.split(topology_file)[0]

            # run destroy command
            args = ["containerlab", "destroy", "-t", topology_file]
            ret = self.run_containerlab_command(
                args=args,
                cwd=topology_folder,
                timeout=timeout,
                ret=ret,
                job=job,
                expect_output=False,
            )

            if not ret.failed:
                ret.result = {lab_name: True}

        return ret

    @Task(fastapi={"methods": ["GET"]})
    def inspect(
        self,
        job: Job,
        lab_name: Union[None, str] = None,
        timeout: int = None,
        details: bool = False,
    ) -> Result:
        """
        Inspect the container lab containers configuration and status.

        This method retrieves information about a specific container lab or all
        container labs, optionally including detailed information.

        Args:
            lab_name (str, optional): The name of the container lab to inspect.
                If not provided, all container labs will be inspected.
            timeout (int, optional): The maximum time in seconds to wait for the
                inspection command to complete. Defaults to None.
            details (bool, optional): Whether to include detailed information in
                the inspection output. Defaults to False.

        Returns:
            Result: An object containing the result of the inspection task.
        """
        timeout = timeout or 600
        ret = Result(task=f"{self.name}:inspect")

        if lab_name:
            args = ["containerlab", "inspect", "-f", "json", "--name", lab_name]
        else:
            args = ["containerlab", "inspect", "-f", "json", "--all"]
        if details:
            args.append("--details")

        ret = self.run_containerlab_command(
            args=args, timeout=timeout, ret=ret, job=job
        )

        # check if lab name given and it is not in output
        if lab_name and lab_name not in ret.result:
            ret.failed = True
            msg = f"'{lab_name}' lab not found"
            ret.errors.append(msg)
            log.error(msg)

        return ret

    @Task(fastapi={"methods": ["POST"]})
    def save(self, job: Job, lab_name: str, timeout: int = None) -> Result:
        """
        Saves the config of a specified lab devices by invoking the `containerlab save` command.

        Args:
            lab_name (str): The name of the lab to save.
            timeout (int, optional): The maximum time in seconds to wait for the operation
                to complete. Defaults to None.

        Returns:
            Result: An object containing the outcome of the save operation. If successful,
                `result` will contain a dictionary with the lab name as the key and `True`
                as the value. If unsuccessful, `failed` will be set to True, and `errors`
                will contain a list of error messages.
        """
        timeout = timeout or 600
        ret = Result(task=f"{self.name}:save")

        # get lab details
        inspect = self.inspect(
            job=job, timeout=timeout, lab_name=lab_name, details=True
        )

        if not inspect.result:
            ret.failed = True
            ret.errors = [f"'{lab_name}' lab not found"]
            ret.result = {lab_name: False}
        else:
            topology_file = inspect.result[lab_name][0]["Labels"]["clab-topo-file"]
            topology_folder = os.path.split(topology_file)[0]

            # run destroy command
            args = ["containerlab", "save", "-t", topology_file]
            ret = self.run_containerlab_command(
                args=args,
                cwd=topology_folder,
                timeout=timeout,
                ret=ret,
                job=job,
                expect_output=False,
            )

            if not ret.failed:
                ret.result = {lab_name: True}

        return ret

    @Task(fastapi={"methods": ["POST"]})
    def restart_lab(self, job: Job, lab_name: str, timeout: int = None) -> Result:
        """
        Restart a specified Containerlab lab.

        This method retrieves the lab details, destroys the existing lab, and redeploys it
        using the provided topology file.

        Args:
            lab_name (str): The name of the lab to restart.
            timeout (int, optional): The timeout value for the operation in seconds. Defaults to None.

        Returns:
            Result: An object containing the status of the operation, any errors encountered,
                    and the result indicating whether the lab was successfully restarted.
        """
        timeout = timeout or 600
        ret = Result(task=f"{self.name}:restart_lab")

        # get lab details
        inspect = self.inspect(
            job=job, timeout=timeout, lab_name=lab_name, details=True
        )

        if not inspect.result:
            ret.failed = True
            ret.errors = [f"'{lab_name}' lab not found"]
            ret.result = {lab_name: False}
        else:
            topology_file = inspect.result[lab_name][0]["Labels"]["clab-topo-file"]
            topology_folder = os.path.split(topology_file)[0]

            # add needed env variables
            env = dict(os.environ)
            env["CLAB_VERSION_CHECK"] = "disable"

            # run destroy command
            args = [
                "containerlab",
                "deploy",
                "-f",
                "json",
                "-t",
                topology_file,
                "--reconfigure",
            ]
            ret = self.run_containerlab_command(
                args=args,
                cwd=topology_folder,
                timeout=timeout,
                ret=ret,
                env=env,
                job=job,
            )

            if not ret.failed:
                ret.result = {lab_name: True}

        return ret

    @Task(fastapi={"methods": ["GET"]})
    def get_nornir_inventory(
        self,
        job: Job,
        lab_name: Union[None, str] = None,
        timeout: int = None,
        groups: Union[None, list] = None,
        use_default_credentials: bool = True,
    ) -> Result:
        """
        Retrieves the Nornir inventory for a specified lab.

        This method inspects the container lab environment and generates a Nornir-compatible
        inventory of hosts based on the lab's configuration. It maps containerlab node kinds
        to Netmiko SSH platform types and extracts relevant connection details.

        Args:
            lab_name (str): The name of the container lab to inspect. If not given loads inventory for all labs.
            timeout (int, optional): The timeout value for the inspection operation. Defaults to None.
            groups (list, optional): A list of group names to assign to the hosts in the inventory. Defaults to None.
            use_default_credentials (bool, optional): Use Containerlab default credentials for hosts or not.

        Returns:
            Result: A `Result` object containing the Nornir inventory. The `result` attribute
            includes a dictionary with host details. If the lab is not found or an error occurs,
            the `failed` attribute is set to True, and the `errors` attribute contains error messages.

        Notes:
            - The method uses a predefined mapping (`norfab.utils.platform_map`) to translate containerlab
              node kinds to Netmiko platform types.
            - If a container's SSH port cannot be determined, it is skipped, and an error is logged.
            - The primary host IP address is determined dynamically using a UDP socket connection or
              by checking the host IP address in the container's port configuration.

        Example of returned inventory structure:
            {
                "hosts": {
                    "host_name": {
                        "hostname": "host_ip",
                        "platform": "netmiko_platform",
                        "groups": ["group1", "group2"],
                    },
                    ...
        """
        timeout = timeout or 600
        groups = groups or []
        ret = Result(task=f"{self.name}:get_nornir_inventory", result={"hosts": {}})

        # get lab details
        inspect = self.inspect(
            job=job, lab_name=lab_name, timeout=timeout, details=True
        )

        # return empty result if lab not found
        if not inspect.result:
            ret.failed = True
            ret.errors = [f"'{lab_name}' lab not found"]
            return ret

        # get host primary IP address
        s = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
        s.setsockopt(socket.SOL_SOCKET, socket.SO_BROADCAST, 1)
        s.connect(("1.2.3.4", 12345))
        primary_host_ip = s.getsockname()[0]
        log.debug(
            f"{self.name} - determined Containerlab host primary IP - '{primary_host_ip}'"
        )

        # form hosts inventory
        for lname, containers in inspect.result.items():
            if lab_name and lname != lab_name:
                continue
            for container in containers:
                host_name = container["Labels"]["clab-node-name"]
                host_port = None
                host_ip = None

                # get ssh port
                for port in container["Ports"]:
                    host_ip = primary_host_ip
                    if port["port"] == 22 and port["protocol"] == "tcp":
                        host_port = port["host_port"]
                        # get host ip address
                        if port["host_ip"] not in [
                            "0.0.0.0",
                            "127.0.0.1",
                            "localhost",
                            "::",
                        ]:
                            host_ip = port["host_ip"]
                        break
                else:
                    log.error(f"{self.name} - {host_name} failed to map SSH port.")
                    continue

                # add host to Nornir inventory
                ret.result["hosts"][host_name] = {
                    "hostname": host_ip,
                    "port": host_port,
                    "groups": groups,
                }

                # get netmiko platform
                clab_platform_name = container["Labels"]["clab-node-kind"]
                netmiko_platform = PlatformMap.convert(
                    "containerlab", "netmiko", clab_platform_name
                )
                if netmiko_platform:
                    ret.result["hosts"][host_name]["platform"] = netmiko_platform[
                        "platform"
                    ]
                else:
                    log.warning(
                        f"{self.name} - {host_name} clab-node-kind '{clab_platform_name}' not mapped to Netmiko platform."
                    )
                    continue

                # get default credentials
                if use_default_credentials:
                    clab_platform = PlatformMap.get("containerlab", clab_platform_name)
                    if not clab_platform:
                        log.warning(
                            f"{self.name} - {host_name} clab-node-kind '{clab_platform_name}' not found."
                        )
                        continue
                    if clab_platform.get("username"):
                        ret.result["hosts"][host_name]["username"] = clab_platform[
                            "username"
                        ]
                    if clab_platform.get("password"):
                        ret.result["hosts"][host_name]["password"] = clab_platform[
                            "password"
                        ]

        return ret

    @Task(fastapi={"methods": ["GET"]})
    def deploy_netbox(
        self,
        job: Job,
        lab_name: str = None,
        tenant: str = None,
        filters: Union[None, list] = None,
        devices: Union[None, list] = None,
        instance: str = None,
        image: str = None,
        ipv4_subnet: str = "172.100.100.0/24",
        ports: tuple = (12000, 15000),
        progress: bool = False,
        reconfigure: bool = False,
        timeout: int = 600,
        node_filter: Union[None, str] = None,
        dry_run: bool = False,
    ) -> Result:
        """
        Deploys a containerlab topology using device data from the Netbox database.

        This method orchestrates the deployment of a containerlab topology by:

        - Inspecting existing containers to determine subnets and ports in use.
        - Allocating a management IPv4 subnet for the new lab, avoiding conflicts.
        - Downloading inventory data from Netbox for the specified lab and filters.
        - Saving the generated topology file to a dedicated folder.
        - Executing the `containerlab deploy` command with appropriate arguments.

        To retrieve topology data from Netbox at least one of these arguments must be provided
        to identify a set of devices to include into Containerlab topology:

        - `tenant` - deploys lab using all devices and links that belong to this tenant
        - `devices` - lab deployed only using devices in the lists
        - `filters` - list of device filters to retrieve from Netbox

        If multiple of above arguments provided, resulting lab topology is a sum of all
        devices matched.

        Args:
            lab_name (str, optional): The name to use for the lab to deploy.
            tenant (str, optional): Deploy lab for given tenant, lab name if not set
                becomes equal to tenant name.
            filters (list, optional): List of filters to apply when fetching devices from Netbox.
            devices (list, optional): List of specific devices to include in the topology.
            instance (str, optional): Netbox instance identifier.
            image (str, optional): Container image to use for devices.
            ipv4_subnet (str, optional): Management IPv4 subnet for the lab.
            ports (tuple, optional): Tuple specifying the range of ports to allocate.
            progress (bool, optional): If True, emits progress events.
            reconfigure (bool, optional): If True, reconfigures an already deployed lab.
            timeout (int, optional): Timeout in seconds for the deployment process.
            node_filter (str, optional): Comma-separated string of nodes to deploy.
            dry_run (bool, optional): If True, only generates and returns the topology
                inventory without deploying.

        Returns:
            Result: deployment results with a list of nodes deployed

        Raises:
            Exception: If the topology file cannot be fetched.
        """
        timeout = timeout or 600
        ret = Result(task=f"{self.name}:deploy_netbox")
        subnets_in_use = set()
        ports_in_use = {}

        # handle lab name and tenant name
        if lab_name is None and tenant:
            lab_name = tenant

        # inspect existing containers
        job.event("Checking existing containers")
        get_containers = self.inspect(job=job, details=True)
        if get_containers.failed is True:
            get_containers.task = f"{self.name}:deploy_netbox"
            return get_containers

        # collect TCP/UDP ports and subnets in use
        job.event("Existing containers found, retrieving details")
        for lname, containers in get_containers.result.items():
            for container in containers:
                clab_name = container["Labels"]["containerlab"]
                clab_topo = container["Labels"]["clab-topo-file"]
                node_name = container["Labels"]["clab-node-name"]
                # collect ports that are in use
                ports_in_use[node_name] = list(
                    set(
                        [
                            f"{p['host_port']}:{p['port']}/{p['protocol']}"
                            for p in container["Ports"]
                            if "host_port" in p and "port" in p and "protocol" in p
                        ]
                    )
                )
                # check existing subnets
                if (
                    container["NetworkSettings"]["IPv4addr"]
                    and container["NetworkSettings"]["IPv4pLen"]
                ):
                    ip = ipaddress.ip_interface(
                        f"{container['NetworkSettings']['IPv4addr']}/"
                        f"{container['NetworkSettings']['IPv4pLen']}"
                    )
                    subnet = str(ip.network.with_prefixlen)
                else:
                    with open(clab_topo, encoding="utf-8") as f:
                        clab_topo_data = yaml.safe_load(f.read())
                        if clab_topo_data.get("mgmt", {}).get("ipv4-subnet"):
                            subnet = clab_topo_data["mgmt"]["ipv4-subnet"]
                        else:
                            msg = f"{clab_name} lab {node_name} node failed to determine mgmt subnet"
                            log.warning(msg)
                            job.event(msg, severity="WARNING")
                            continue
                subnets_in_use.add(subnet)
                # reuse existing lab subnet
                if clab_name == lab_name:
                    ipv4_subnet = subnet
                    job.event(
                        f"{ipv4_subnet} not in use by existing containers, using it"
                    )
                # allocate new subnet if its in use by other lab
                elif clab_name != lab_name and ipv4_subnet == subnet:
                    msg = f"{ipv4_subnet} already in use, allocating new subnet"
                    log.info(msg)
                    job.event(msg)
                    ipv4_subnet = None

        job.event("Collected TCP/UDP ports used by existing containers")

        # allocate new subnet
        if ipv4_subnet is None:
            pool = set(f"172.100.{i}.0/24" for i in range(100, 255))
            ipv4_subnet = list(sorted(pool.difference(subnets_in_use)))[0]
            job.event(f"{lab_name} allocated new subnet {ipv4_subnet}")

        job.event(f"{lab_name} fetching lab topology data from Netbox")

        # download inventory data from Netbox
        netbox_reply = self.client.run_job(
            service="netbox",
            task="get_containerlab_inventory",
            workers="any",
            timeout=timeout,
            kwargs={
                "lab_name": lab_name,
                "tenant": tenant,
                "filters": filters,
                "devices": devices,
                "instance": instance,
                "image": image,
                "ipv4_subnet": ipv4_subnet,
                "ports": ports,
                "ports_map": ports_in_use,
                "progress": progress,
            },
        )

        # use inventory from first worker that returned hosts data
        for wname, wdata in netbox_reply.items():
            if wdata["failed"] is False and wdata["result"]:
                topology_inventory = wdata["result"]
                break
        else:
            msg = f"{self.name} - Netbox returned no data for '{lab_name}' lab"
            log.error(msg)
            raise RuntimeError(msg)

        job.event(f"{lab_name} topology data retrieved from Netbox")

        if dry_run is True:
            ret.result = topology_inventory
            return ret

        # create folder to store topology
        topology_folder = os.path.join(self.topologies_dir, lab_name)
        os.makedirs(topology_folder, exist_ok=True)

        # create topology file
        topology_file = os.path.join(topology_folder, f"{lab_name}.yaml")
        with open(topology_file, "w", encoding="utf-8") as tf:
            tf.write(yaml.dump(topology_inventory, default_flow_style=False))

        job.event(f"{lab_name} topology data saved to '{topology_file}'")

        return self.deploy(
            job=job,
            topology=topology_file,
            reconfigure=reconfigure,
            timeout=timeout,
            node_filter=node_filter or "",
        )



"""
Collection of common text processing utils
"""

import re


def slugify(value: str) -> str:
    """Convert a string to Django slug format"""
    value = str(value).strip()
    value = re.sub(r"[^\w\s-]", "", value.lower())
    return re.sub(r"[-\s]+", "-", value).strip("-_")



import logging

from typing import Union, Any
from norfab.models import Result
from norfab.core.worker import Task, Job
from norfab.core.exceptions import UnsupportedPluginError
from nornir_salt.plugins.tasks import (
    netmiko_send_commands,
    scrapli_send_commands,
    napalm_send_commands,
    nr_test,
)
from nornir_salt.plugins.functions import ResultSerializer
from norfab.clients.shell_clients.nornir.nornir_picle_shell_cli import NorniCliInput

log = logging.getLogger(__name__)


class CliTask:
    @Task(
        fastapi={"methods": ["POST"]},
        input=NorniCliInput,
    )
    def cli(
        self,
        job: Job,
        commands: Union[str, list] = None,
        plugin: str = "netmiko",
        dry_run: bool = False,
        run_ttp: str = None,
        job_data: Any = None,
        to_dict: bool = True,
        add_details: bool = False,
        **kwargs: Any,
    ) -> Result:
        """
        Task to collect/retrieve show commands output from network devices using
        Command Line Interface (CLI).

        Must either provide list of commands to run or TTP template to run.

        Args:
            job: NorFab Job object containing relevant metadata
            commands (list, optional): List of commands to send to devices or URL to a file or template
                URL that resolves to a file.
            plugin (str, optional): Plugin name to use. Valid options are
                ``netmiko``, ``scrapli``, ``napalm``.
            dry_run (bool, optional): If True, do not send commands to devices,
                just return them.
            run_ttp (str, optional): TTP Template to run.
            job_data (str, optional): URL to YAML file with data or dictionary/list
                of data to pass on to Jinja2 rendering context.
            to_dict (bool, optional): If True, returns results as a dictionary.
            add_details (bool, optional): If True, adds task execution details
                to the results.
            **kwargs: Additional arguments to pass to the specified task plugin.

        Returns:
            dict: A dictionary with the results of the CLI task.

        Raises:
            UnsupportedPluginError: If the specified plugin is not supported.
            FileNotFoundError: If the specified TTP template or job data file
                cannot be downloaded.
        """
        job_data = job_data or {}
        timeout = job.timeout * 0.9
        ret = Result(task=f"{self.name}:cli", result={} if to_dict else [])

        filtered_nornir, no_match_result = self.filter_hosts_and_validate(kwargs, ret)
        if ret.status == "no_match":
            return ret

        # decide on what send commands task plugin to use
        if plugin == "netmiko":
            task_plugin = netmiko_send_commands
            if kwargs.get("use_ps"):
                kwargs.setdefault("timeout", timeout)
            else:
                kwargs.setdefault("read_timeout", timeout)
        elif plugin == "scrapli":
            task_plugin = scrapli_send_commands
            kwargs.setdefault("timeout_ops", timeout)
        elif plugin == "napalm":
            task_plugin = napalm_send_commands
        else:
            raise UnsupportedPluginError(f"Plugin '{plugin}' not supported")

        # download TTP template
        if self.is_url(run_ttp):
            downloaded = self.fetch_file(run_ttp)
            kwargs["run_ttp"] = downloaded
            if downloaded is None:
                msg = f"{self.name} - TTP template download failed '{run_ttp}'"
                raise FileNotFoundError(msg)
        # use TTP template as is - inline template or ttp://xyz path
        elif run_ttp:
            kwargs["run_ttp"] = run_ttp

        # download job data
        job_data = self.load_job_data(job_data)

        nr = self._add_processors(filtered_nornir, kwargs, job)  # add processors

        # render commands using Jinja2 on a per-host basis
        if commands:
            commands = commands if isinstance(commands, list) else [commands]
            for host in nr.inventory.hosts.values():
                rendered = self.jinja2_render_templates(
                    templates=commands,
                    context={
                        "host": host,
                        "norfab": self.client,
                        "job_data": job_data,
                        "netbox": self.add_jinja2_netbox(),
                    },
                    filters=self.add_jinja2_filters(),
                )
                host.data["__task__"] = {"commands": rendered}

        # run task
        log.debug(
            f"{self.name} - running cli commands '{commands}', kwargs '{kwargs}', is cli dry run - '{dry_run}'"
        )
        if dry_run is True:
            result = nr.run(
                task=nr_test, use_task_data="commands", name="dry_run", **kwargs
            )
        else:
            with self.connections_lock:
                result = nr.run(task=task_plugin, **kwargs)

        ret.failed = result.failed  # failed is true if any of the hosts failed
        ret.result = ResultSerializer(result, to_dict=to_dict, add_details=add_details)

        # remove __task__ data
        for host_name, host_object in nr.inventory.hosts.items():
            _ = host_object.data.pop("__task__", None)

        self.watchdog.connections_update(nr, plugin)
        self.watchdog.connections_clean()

        return ret
